
/**************************************************************************
 *
 *  $Id: mbgdemo.c 2.4 2005/03/16 14:56:37 martin REL_M $
 *
 *  Description:
 *    Demo program which shows how to access Meinberg radio clock devices
 *    using the mbgdevio library.
 *    Prototypes of available API functions can be found in mbgdevio.h.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbgdemo.c $
 *  Revision 2.4  2005/03/16 14:56:37  martin
 *  Increased version number due to updated driver library.
 *  Modified copyright year.
 *  Revision 2.3  2004/11/10 10:43:25  martin
 *  Added some type casts to prevent from compiler warnings.
 *  Revision 2.2  2004/03/10 17:28:24  martin
 *  Changes due to renamed symbols in the library.
 *  Revision 2.1  2003/05/16 10:01:52  martin
 *  Initial version using the mbgdevio library.
 *
 **************************************************************************/

// Meinberg headers
#include <mbgdevio.h>
#include <gpsutils.h>

// Standard headers
#include <stdio.h>


static const char *pname = "mbgdemo";
static const char *pversion = "v2.4";
static const char *pcopyright = "(c) Meinberg 2001-2005";



static /*HDR*/
void err_msg( const char *msg, int rc )
{
  fprintf( stderr, "** %s: %i\n", msg, rc );

}  // err_msg



#if !defined _PCPSDRVR_H

// Driver info does only make sense if a user space program
// accesses the hardware via a kernel device driver.

static /*HDR*/
void print_drvr_info( void )
{
  int rc;
  PCPS_DRVR_INFO drvr_info;
  MBG_DEV_HANDLE dh = mbg_open_device( 0 );

  if ( dh == MBG_INVALID_DEV_HANDLE )
  {
    perror( "Unable to open device" );
    exit( 1 );
  }


  rc = mbg_get_drvr_info( dh, &drvr_info );

  mbg_close_device( &dh );


  if ( rc != PCPS_SUCCESS )
  {
    err_msg( "Failed to read driver info", rc );
    exit( 1 );
  }

  printf( "Kernel driver: %s v%i.%02i, %i device%s\n\n",
          drvr_info.id_str,
          drvr_info.ver_num / 100,
          drvr_info.ver_num % 100,
          drvr_info.n_devs,
          ( drvr_info.n_devs == 1 ) ? "" : "s"
        );

}  // print_drvr_info

#endif



static /*HDR*/
void print_dev_info( MBG_DEV_HANDLE dh,
                     PCPS_DEV *p_dev )
{
  int rc;

  rc = mbg_get_device_info( dh, p_dev );

  if ( rc == PCPS_SUCCESS )
  {
    printf( "  %s at port %03Xh\n",
            _pcps_type_name( p_dev ),
            _pcps_port_base( p_dev, 0 )
          );
  }
  else
    err_msg( "Failed to read device info", rc );

}  // print_dev_info



static /*HDR*/
void print_date_time( PCPS_TIME *tp, const char *label )
{
  if ( label )
    printf( label );

  printf( "%02u.%02u.%02u  %02u:%02u:%02u.%02u (UTC%+dh)\n",
          tp->mday, tp->month, tp->year,
          tp->hour, tp->min, tp->sec, tp->sec100,
          tp->offs_utc
        );

}  // print_date_time



static /*HDR*/
void print_date_time_status( MBG_DEV_HANDLE dh )
{
  PCPS_TIME t;
  int rc = mbg_get_time( dh, &t );

  if ( rc == PCPS_SUCCESS )
  {
    print_date_time( &t, "  " );

    printf( "  %s\n",
            ( t.status & PCPS_FREER ) ?
            "free running" :
            "synchronized" );

    printf( "  %s\n",
            ( t.status & PCPS_SYNCD ) ?
            "synchronized after last RESET" :
            "not synchronized after last RESET" );
  }
  else
    err_msg( "Failed to read date/time/status", rc );

}  // print_date_time_status



static /*HDR*/
void print_hr_time( MBG_DEV_HANDLE dh )
{
  PCPS_HR_TIME hr_t;
  int rc = mbg_get_hr_time( dh, &hr_t );

  if ( rc == PCPS_SUCCESS )
  {
    printf( "  HR time: %08lX.%08lX UTC%+ldsec\n",
            (ulong) hr_t.tstamp.sec,
            (ulong) hr_t.tstamp.frac,
            (long) hr_t.utc_offs
          );
  }
  else
    err_msg( "Failed to read High Resolution Time", rc );

}  // print_hr_time



static /*HDR*/
void print_receiver_position( MBG_DEV_HANDLE dh )
{
  POS pos;
  char ws[256];
  int rc = mbg_get_gps_pos( dh, &pos );

  if ( rc == PCPS_SUCCESS )
  {
    sprint_pos_geo( ws, &pos, ", ", 4 );
    printf( "  Receiver position: %s\n", ws );
  }
  else
    err_msg( "Failed to read receiver position", rc );

}  // print_receiver_position



static /*HDR*/
void print_sv_info( MBG_DEV_HANDLE dh )
{
  STAT_INFO stat_info;
  int rc = mbg_get_gps_stat_info( dh, &stat_info );


  if ( rc == PCPS_SUCCESS )
  {
    printf( "  Satellites: %u in view, %u good\n",
            stat_info.svs_in_view,
            stat_info.good_svs
          );
  }
  else
    err_msg( "Failed to read GPS status info", rc );

}  // print_sv_info



static /*HDR*/
void check_user_captures( MBG_DEV_HANDLE dh, PCPS_DEV *pdev )
{
  TTM ucap;
  int rc;
  unsigned int ucaps_read = 0;


  for (;;)   // read all entries from capture buffer
  {
    rc = mbg_get_gps_ucap( dh, &ucap );

    if ( rc != PCPS_SUCCESS )
    {
      err_msg( "Failed to read user captures", rc );
      break;
    }


    // If a user capture event has been read
    // then it it removed from the clock's buffer.

    // If no new capture is available, the ucap.tm structure
    // is set to "unread".
    if ( !_pcps_time_is_read( &ucap.tm ) ) // no new user capture entry
    {
      if ( ucaps_read )
        printf( "  User captures read: %u\n", ucaps_read );
      else
        printf( "  No user captures to be read.\n" );

      break;
    }

    printf( "  New capture: CH%i: %02i.%02i.%02i  %2i:%02i:%02i.%07li\n",
            ucap.channel,
            ucap.tm.mday,
            ucap.tm.month,
            ucap.tm.year % 100,
            ucap.tm.hour,
            ucap.tm.min,
            ucap.tm.sec,
            (long) ucap.tm.frac
          );

    ucaps_read++;
  }


  if ( _pcps_can_clr_ucap_buff( pdev ) )
  {
    rc = mbg_clr_ucap_buff( dh );

    if ( rc == PCPS_SUCCESS )
      printf( "  Capture buffer cleared.\n" );
    else
      err_msg( "  Failed to clear capture buffer", rc );
  }
  else
    printf( "  Clearing capture buffer not supported.\n" );

}  // check_user_captures



int main( int argc, char* argv[] )
{
  int i;
  int rc = 0;
  int devices_found;


  printf( "\n%s %s %s\n\n", pname, pversion, pcopyright );

  devices_found = mbg_find_devices();

  if ( devices_found == 0 )
  {
    printf( "No radio clock found.\n\n" );
    return 1;
  }


  printf( "Found %i radio clock%s.\n\n",
          devices_found,
          ( devices_found == 1 ) ? "" : "s"
        );


  #if !defined( _PCPSDRVR_H )
    print_drvr_info();
  #endif


  // There may be several radio clock devices installed.
  // Try to get information from each of the devices.
  for ( i = 0; i < devices_found; i++ )
  {
    static PCPS_DEV dev;

    MBG_DEV_HANDLE dh;


    printf( "Radio clock %i:\n", i );

    dh = mbg_open_device( i );

    if ( dh == MBG_INVALID_DEV_HANDLE )
    {
      err_msg( "Unable to open device", rc );
      continue;
    }


    print_dev_info( dh, &dev );

    // Read the current date, time, and status.
    print_date_time_status( dh );

    // Some newer clocks (currently GPS only) support
    // a high resolution time. If supported, read and
    // display the HR time.
    if ( _pcps_has_hr_time( &dev ) )
      print_hr_time( dh );

    // Print some GPS clock specific info
    if ( _pcps_is_gps( &dev ) )
    {
      print_receiver_position( dh );
      print_sv_info( dh );
      check_user_captures( dh, &dev );
    }

    mbg_close_device( &dh );

    printf( "\n" );
  }

  printf ( "\n" );

  return rc;
}
