
/**************************************************************************
 *
 *  $Id: mbgdevio.c 1.17 2006/05/02 13:15:37 martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Functions called from user space to access Meinberg device drivers.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbgdevio.c $
 *  Revision 1.17  2006/05/02 13:15:37  martin
 *  Added mbg_set_gps_port_settings(), mbg_get_gps_all_pout_info(),
 *  mbg_set_gps_pout_settings_idx(), mbg_set_gps_pout_settings().
 *  Revision 1.16  2005/06/02 11:53:12Z  martin
 *  Implemented existing mbg_generic_..() functions.
 *  Added new functions mbg_generic_io() and mbg_dev_has_generic_io().
 *  Added new function mbg_get_synth_state().
 *  Partially use inline function.
 *  Changed order of some functions.
 *  More simplifications using macros.
 *  Unified macros for Win32 and Linux.
 *  Fixed warning under Win32 using type cast.
 *  Revision 1.15  2005/01/31 16:44:21Z  martin
 *  Added function mbg_get_hr_time_comp() which returns HR time stamp 
 *  which has latency compensated.
 *  Revision 1.14  2005/01/14 10:22:23Z  martin
 *  Added functions which query device features.
 *  Revision 1.13  2004/12/09 11:23:59Z  martin
 *  Support configuration of on-board frequency synthesizer.
 *  Revision 1.12  2004/11/09 14:11:07Z  martin
 *  Modifications were required in order to be able to configure IRIG 
 *  settings of cards which provide both IRIG input and output.
 *  Renamed functions mbg_get_irig_info() and mbg_set_irig_settings() 
 *  to mbg_get_irig_rx_info() and mbg_set_irig_rx_settings() 
 *  New functions mbg_get_irig_tx_info() and mbg_set_irig_tx_settings().
 *  All API functions now use well defined parameter types instead of
 *  generic types. Some new types have been defined therefore.
 *  Added a workaround for GPS169PCI cards with early firmware versions
 *  which used the same codes to configure the IRIG output as the TCR 
 *  cards use to configure the IRIG input. Those codes are now 
 *  exclusively used to configure the IRIG input. The workaround
 *  has been included in order to let GPS169PCI cards work properly 
 *  after a driver update, without requiring a firmware update.
 *  The macro _pcps_ddev_requires_irig_workaround() is used to check 
 *  if the workaround is required.
 *  Renamed function mbg_get_gps_stat() to mbg_get_gps_bvar_stat().
 *  Revision 1.11  2004/08/17 11:13:45Z  martin
 *  Account for renamed symbols.
 *  Revision 1.10  2004/04/14 09:39:17Z  martin
 *  Allow [g|s]et_irig_info() also for new devices with IRIG output.
 *  Use MBGDEVIO_COMPAT_VERSION to check version.
 *  Revision 1.9  2003/12/22 15:30:52Z  martin
 *  Added functions mbg_get_asic_version(), mbg_get_time_cycles(),
 *  and mbg_get_hr_time_cycles().
 *  Support higher baud rates for TCR510PCI and PCI510.
 *  Support PCPS_HR_TIME for TCR510PCI.
 *  API calls return ioctl results instead of success/-1.
 *  Moved some Win32 specific code to mbgsvctl DLL.
 *  Log Win32 ioctl errors to event log for debugging.
 *  Revision 1.8  2003/06/19 08:42:33Z  martin
 *  Renamed function mbg_clr_cap_buff() to mbg_clr_ucap_buff().
 *  New functions mbg_get_ucap_entries() and mbg_get_ucap_event().
 *  New function mbg_get_hr_ucap().
 *  New functions mbgdevio_get_version() and mbgdevio_check_version().
 *  New functions for generic read/write access.
 *  New functions mbg_get_pcps_tzdl() and mbg_set_pcps_tzdl().
 *  Fixed a bug passing the wrong command code to a 
 *  direct access target in mbg_get_sync_time().
 *  Return driver info for direct access targets.
 *  Include pcpsdrvr.h and pcps_dos.h, if applicable.
 *  For direct access targets, check if a function is supported
 *  before accessing the hardware.
 *  Use const parameter pointers if applicable.
 *  Changes due to renamed symbols/macros.
 *  Source code cleanup.
 *  Revision 1.7  2003/05/16 08:52:46  MARTIN
 *  Swap doubles inside API functions.
 *  Enhanced support for direct access targets.
 *  Removed obsolete code.
 *  Revision 1.6  2003/04/25 10:14:16  martin
 *  Renamed macros.
 *  Extended macro calls for direct access targets.
 *  Updated macros for Linux.
 *  Revision 1.5  2003/04/15 19:35:25Z  martin
 *  New functions mbg_setup_receiver_info(), 
 *  mbg_get_serial_settings(), mbg_save_serial_settings().
 *  Revision 1.4  2003/04/09 16:07:16Z  martin
 *  New API functions mostly complete.
 *  Use renamed IOCTL codes from mbgioctl.h.
 *  Added DllEntry function foe Win32.
 *  Made MBG_Device_count and MBG_Device_Path local.
 *  Revision 1.3  2003/01/24 13:44:40Z  martin
 *  Fixed get_ref_time_from_driver_at_sec_change() to be used 
 *  with old kernel drivers.
 *  Revision 1.2  2002/09/06 11:04:01Z  martin
 *  Some old API functions have been replaced by new ones
 *  for a common PnP/non-PnP API.
 *  New API function which clears capture buffer.
 *  New function get_ref_time_from_driver_at_sec_change().
 *  Revision 1.1  2002/02/19 13:48:20Z  MARTIN
 *  Initial revision
 *
 **************************************************************************/

#define _MBGDEVIO
 #include <mbgdevio.h>
#undef _MBGDEVIO

#include <parmpcps.h>
#include <parmgps.h>
#include <gpsutils.h>


#if defined( MBG_USE_KERNEL_DRIVER )

  #include <mbgioctl.h>

#else

  #include <pcpsdrvr.h>
  #include <pci_asic.h>

  static PCPS_DRVR_INFO drvr_info = { MBGDEVIO_VERSION, 0, "MBGDEVIO direct" };

#endif



// target specific code for different environments

#if defined( MBG_TGT_WIN32 )

  #include <mbgsvctl.h>
  #include <tchar.h>
  #include <stdio.h>


  #define __mbg_inline __forceinline

  #define _mbgdevio_vars() \
    DWORD rc

  #define _mbgdevio_ret_val \
    rc

  #define _do_mbgdevio_io( _dh, _ioctl, _p, _insz, _outsz ) \
      rc = do_mbgdevio_io( _dh, _ioctl, (LPVOID) _p, _insz, _outsz )


  static __mbg_inline
  DWORD do_mbgdevio_io( MBG_DEV_HANDLE dh, int ioctl_code, 
                        LPVOID p, int in_sz, int out_sz )
  {
    DWORD ReturnedLength;

    if ( !DeviceIoControl( dh, ioctl_code,
                           p, in_sz, p, out_sz,
                           &ReturnedLength,
                           NULL
                         ) )
    {
      int rc = GetLastError();
      mbgsvctl_log_mbgdevio_error( ioctl_code, rc );
      return rc;
    }

    return PCPS_SUCCESS;

  }  // do_mbgdevio_io

#elif defined( MBG_TGT_LINUX )

  #include <mbgioctl.h>

  #include <unistd.h>
  #include <stdlib.h>
  #include <string.h>
  #include <fcntl.h>
  #include <sys/ioctl.h>


  #define __mbg_inline __inline__

  #define _mbgdevio_vars() \
    int rc

  #define _mbgdevio_ret_val \
    ( ( rc < 0 ) ? rc : PCPS_SUCCESS )

  #define _do_mbgdevio_io( _dh, _ioctl, _p, _insz, _outsz ) \
    rc = ioctl( _dh, _ioctl, _p )
                               
#else // other target OSs which access the hardware directly

  #if defined MBG_TGT_QNX_NTO
    #include <stdio.h>
    #include <sys/neutrino.h>
  #endif

  #if MBG_USE_DOS_TSR
    #include <pcps_dos.h>
  #else
    #define pcps_read_safe            _pcps_read
    #define pcps_write_safe           pcps_write
    #define pcps_read_gps_safe        pcps_read_gps
    #define pcps_write_gps_safe       pcps_write_gps

    #define _pcps_write_byte_safe     _pcps_write_byte
    #define _pcps_read_var_safe       _pcps_read_var
    #define _pcps_write_var_safe      _pcps_write_var
    #define _pcps_read_gps_var_safe   _pcps_read_gps_var
    #define _pcps_write_gps_var_safe  _pcps_write_gps_var
  #endif


  #define _mbgdevio_vars() \
    int rc

  #define _mbgdevio_ret_val \
    rc

  #define _mbgdevio_chk_cond( _cond ) \
  {                                   \
    if ( !(_cond) )                   \
      return -1;                      \
  }                                   \

#endif  // end of target specific code



#if !defined( _mbgdevio_chk_cond )
  // If the macro has not been defined previously then 
  // it is not be required for the target environment and
  // is defined as empty string.
  #define _mbgdevio_chk_cond( _cond );
#endif



// The code below depends on whether the target device is accessed via
// IOCTLs to a device driver, or the hardware is accessed directly.

#if defined( _MBGIOCTL_H )  // using IOCTL to access device driver

  #define _mbgdevio_read( _dh, _cmd, _ioctl, _p, _sz ) \
    _do_mbgdevio_io( _dh, _ioctl, _p, 0, _sz )

  #define _mbgdevio_write( _dh, _cmd, _ioctl, _p, _sz ) \
    _do_mbgdevio_io( _dh, _ioctl, _p, _sz, 0 )

  #define _mbgdevio_read_gps    _mbgdevio_read

  #define _mbgdevio_write_gps   _mbgdevio_write


  #define _mbgdevio_read_var( _dh, _cmd, _ioctl, _p ) \
    _mbgdevio_read( _dh, _cmd, _ioctl, _p, sizeof( *(_p) ) )

  #define _mbgdevio_write_var( _dh, _cmd, _ioctl, _p ) \
    _mbgdevio_write( _dh, _cmd, _ioctl, _p, sizeof( *(_p) ) )

  #define _mbgdevio_write_cmd( _dh, _cmd, _ioctl ) \
    _mbgdevio_write( _dh, _cmd, _ioctl, NULL, 0 )

  #define _mbgdevio_read_gps_var    _mbgdevio_read_var

  #define _mbgdevio_write_gps_var   _mbgdevio_write_var


  #define _mbgdevio_gen_read( _dh, _cmd, _p, _sz ) \
    rc = mbgdevio_do_gen_io( _dh, _cmd, IOCTL_PCPS_GENERIC_READ, NULL, 0, _p, _sz )

  #define _mbgdevio_gen_write( _dh, _cmd, _p, _sz ) \
    rc = mbgdevio_do_gen_io( _dh, _cmd, IOCTL_PCPS_GENERIC_WRITE, _p, _sz, NULL, 0 )

  #define _mbgdevio_gen_io( _dh, _type, _in_p, _in_sz, _out_p, _out_sz ) \
    rc = mbgdevio_do_gen_io( _dh, _type, IOCTL_PCPS_GENERIC_IO, _in_p, _in_sz, _out_p, _out_sz )

  #define _mbgdevio_gen_read_gps( _dh, _cmd, _p, _sz ) \
    rc = mbgdevio_do_gen_io( _dh, _cmd, IOCTL_PCPS_GENERIC_READ_GPS, NULL, 0, _p, _sz )

  #define _mbgdevio_gen_write_gps( _dh, _cmd, _p, _sz ) \
    rc = mbgdevio_do_gen_io( _dh, _cmd, IOCTL_PCPS_GENERIC_WRITE_GPS, _p, _sz, NULL, 0 )


  static __mbg_inline
  int mbgdevio_do_gen_io( MBG_DEV_HANDLE dh, int info, int ioctl_code, 
                           const void *in_p, int in_sz, 
                           void *out_p, int out_sz )
  {
    IOCTL_GENERIC_BUFFER *p_buff;
    int buff_size = sizeof( p_buff->ctl )
                + ( ( in_sz > out_sz ) ? in_sz : out_sz );
    int rc;
                    
    p_buff = (IOCTL_GENERIC_BUFFER *) malloc( buff_size );
    
    if ( p_buff == NULL )
      return -1;

    p_buff->ctl.info = info;
    p_buff->ctl.data_size_in = in_sz;
    p_buff->ctl.data_size_out = out_sz;

    if ( in_p )
      memcpy( p_buff->data, in_p, in_sz );

    _do_mbgdevio_io( dh, ioctl_code,
                     p_buff,
                     sizeof( IOCTL_GENERIC_CTL ) + in_sz,
                     sizeof( IOCTL_GENERIC_CTL ) + out_sz );

    if ( out_p && ( rc == PCPS_SUCCESS ) )
      memcpy( out_p, p_buff->data, out_sz );
      
    free( p_buff );

    return rc;

  }  // mbgdevio_do_gen_io

#else  // accessing hardware device directly

  #define _mbgdevio_read( _dh, _cmd, _ioctl, _p, _sz ) \
    rc = pcps_read_safe( _dh, _cmd, _p, _sz )

  #define _mbgdevio_write( _dh, _cmd, _ioctl, _p, _sz ) \
    rc = pcps_write_safe( _dh, _cmd, _p, _sz )

  #define _mbgdevio_read_gps( _dh, _cmd, _ioctl, _p, _sz ) \
    rc = pcps_read_gps_safe( _dh, _cmd, _p, _sz )

  #define _mbgdevio_write_gps( _dh, _cmd, _ioctl, _p, _sz ) \
    rc = pcps_write_gps_safe( _dh, _cmd, _p, _sz )


  #define _mbgdevio_read_var( _dh, _cmd, _ioctl, _p ) \
    rc = _pcps_read_var_safe( _dh, _cmd, *(_p) )

  #define _mbgdevio_write_var( _dh, _cmd, _ioctl, _p ) \
    rc = _pcps_write_var_safe( _dh, _cmd, *(_p) )

  #define _mbgdevio_write_cmd( _dh, _cmd, _ioctl ) \
    rc = _pcps_write_byte_safe( _dh, _cmd );

  #define _mbgdevio_read_gps_var( _dh, _cmd, _ioctl, _p ) \
    rc = _pcps_read_gps_var_safe( _dh, _cmd, *(_p) )

  #define _mbgdevio_write_gps_var( _dh, _cmd, _ioctl, _p ) \
    rc = _pcps_write_gps_var_safe( _dh, _cmd, *(_p) )


  #define _mbgdevio_gen_read( _dh, _cmd, _p, _sz ) \
    _mbgdevio_read( _dh, _cmd, -1, _p, _sz )

  #define _mbgdevio_gen_write( _dh, _cmd, _p, _sz ) \
    _mbgdevio_write( _dh, _cmd, -1, _p, _sz )

  #define _mbgdevio_gen_io( _dh, _type, _in_p, _in_sz, _out_p, _out_sz ); \
  {                                                                       \
    _mbgdevio_chk_cond( _pcps_ddev_has_generic_io( _dh ) );               \
    rc = pcps_generic_io( _dh, _type, _in_p, _in_sz, _out_p, _out_sz );   \
  }

  #define _mbgdevio_gen_read_gps( _dh, _cmd, _p, _sz ) \
    _mbgdevio_read_gps( _dh, _cmd, -1, _p, _sz )

  #define _mbgdevio_gen_write_gps( _dh, _cmd, _p, _sz ) \
    _mbgdevio_write_gps( _dh, _cmd, -1, _p, _sz )

#endif




#define _mbgdevio_read_chk( _dh, _cmd, _ioctl, _p, _sz, _cond ) \
{                                                               \
  _mbgdevio_chk_cond( _cond );                                  \
  _mbgdevio_read( _dh, _cmd, _ioctl, _p, _sz );                 \
}

#define _mbgdevio_read_var_chk( _dh, _cmd, _ioctl, _p, _cond ) \
{                                                              \
  _mbgdevio_chk_cond( _cond );                                 \
  _mbgdevio_read_var( _dh, _cmd, _ioctl, _p );                 \
}

#define _mbgdevio_write_var_chk( _dh, _cmd, _ioctl, _p, _cond ) \
{                                                               \
  _mbgdevio_chk_cond( _cond );                                  \
  _mbgdevio_write_var( _dh, _cmd, _ioctl, _p );                 \
}

#define _mbgdevio_write_cmd_chk( _dh, _cmd, _ioctl, _cond ) \
{                                                           \
  _mbgdevio_chk_cond( _cond );                              \
  _mbgdevio_write_cmd( _dh, _cmd, _ioctl );                 \
}

#define _mbgdevio_read_gps_chk( _dh, _cmd, _ioctl, _p, _sz, _cond ) \
{                                                                   \
  _mbgdevio_chk_cond( _cond );                                      \
  _mbgdevio_read_gps( _dh, _cmd, _ioctl, _p, _sz );                 \
}

#define _mbgdevio_read_gps_var_chk( _dh, _cmd, _ioctl, _p, _cond ) \
{                                                                  \
  _mbgdevio_chk_cond( _cond );                                     \
  _mbgdevio_read_gps_var( _dh, _cmd, _ioctl, _p );                 \
}

#define _mbgdevio_write_gps_var_chk( _dh, _cmd, _ioctl, _p, _cond ) \
{                                                                   \
  _mbgdevio_chk_cond( _cond );                                      \
  _mbgdevio_write_gps_var( _dh, _cmd, _ioctl, _p );                 \
}



#if defined( _MBGIOCTL_H )
  #define _mbgdevio_query_cond( _dh, _cond, _ioctl, _p ) \
  {                                                      \
    _mbgdevio_vars();                                    \
    _mbgdevio_read_var( _dh, -1, _ioctl, _p );           \
    return _mbgdevio_ret_val;                            \
  }
#else
  #define _mbgdevio_query_cond( _dh, _cond, _ioctl, _p ) \
  {                                                      \
    *p = _cond( _dh );                                   \
    return PCPS_SUCCESS;                                 \
  }
#endif





/*HDR*/
_MBG_API_ATTR int _MBG_API mbgdevio_get_version( void )
{

  return MBGDEVIO_VERSION;

}  // mbgdevio_get_version



/*HDR*/
_MBG_API_ATTR int _MBG_API mbgdevio_check_version( int header_version )
{
  if ( header_version >= MBGDEVIO_COMPAT_VERSION )
    return PCPS_SUCCESS;

  return -1;

}  // mbgdevio_check_version



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_find_devices( void )
{
  #if defined( _PCPSDRVR_H )

    #if defined MBG_TGT_QNX_NTO
      // Since this program accessed the hardware directly
      // I/O privileges must be assigned to the thread.
      if ( ThreadCtl( _NTO_TCTL_IO, NULL ) == -1 )
      {
        perror( "Fatal error" );
        exit( 1 );
      }
    #endif

    pcps_detect_clocks( pcps_isa_ports, NULL );

    return n_ddevs;

  #elif defined( MBG_TGT_WIN32 )

    return mbg_svc_find_devices();

  #elif defined( MBG_TGT_LINUX )

    MBG_DEV_HANDLE dh = mbg_open_device( 0 );

    if ( dh != MBG_INVALID_DEV_HANDLE )
    {
      mbg_close_device( &dh );
      return 1;
    }

    return 0;

  #endif

}  // mbg_find_devices



/*HDR*/
_MBG_API_ATTR MBG_DEV_HANDLE _MBG_API mbg_open_device( unsigned int device_index )
{
#if defined( MBG_TGT_WIN32 )

  const char *device_path;
  HANDLE dh;

  device_path = mbg_svc_get_device_path( device_index );

  if ( device_path == NULL )
    goto fail;


  dh = CreateFile( 
         device_path,                   // file name
         GENERIC_READ | GENERIC_WRITE,  // access mode
         0,                             // share mode
         NULL,                          // security descriptor
         OPEN_EXISTING,                 // how to create
         0,                             // file attributes
         NULL                           // handle to template file
       );

  if ( INVALID_HANDLE_VALUE == dh )
  {
    #if 0 //##++
      printf( "mbg_open_device: CreateFile failed for index %i\n", device_index );
    #endif

    goto fail;
  }

  return dh;

fail:  
  return MBG_INVALID_DEV_HANDLE;

#elif defined ( MBG_TGT_LINUX )

  static const char *dev_fn = "/dev/mbgclk";    //##++

  MBG_DEV_HANDLE dh = open( dev_fn, 0 );

  return ( dh < 0 ) ? MBG_INVALID_DEV_HANDLE : dh;

#else

  return ( device_index < n_ddevs ) ? &pcps_ddev[device_index] : NULL;

#endif

} // mbg_open_device



/*HDR*/
_MBG_API_ATTR void _MBG_API mbg_close_device( MBG_DEV_HANDLE *dev_handle )
{
#if defined( MBG_TGT_WIN32 )

  if ( *dev_handle != MBG_INVALID_DEV_HANDLE )
    CloseHandle( *dev_handle );

#elif defined ( MBG_TGT_LINUX )

  close( *dev_handle );

#endif

 *dev_handle = MBG_INVALID_DEV_HANDLE;

}  // mbg_close_device



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_drvr_info( MBG_DEV_HANDLE dh, PCPS_DRVR_INFO *p )
{
  #if defined( _MBGIOCTL_H )
    _mbgdevio_vars();
    _mbgdevio_read_var( dh, -1, IOCTL_GET_PCPS_DRVR_INFO, p );
    return _mbgdevio_ret_val;
  #else
    #if defined( __BORLANDC__ )
      dh;   // avoid warnings "never used"
    #endif
    drvr_info.n_devs = n_ddevs;
    *p = drvr_info;
    return PCPS_SUCCESS;
  #endif
}  // mbg_get_drvr_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_device_info( MBG_DEV_HANDLE dh, PCPS_DEV *p )
{
  #if defined( _MBGIOCTL_H )
    _mbgdevio_vars();
    _mbgdevio_read_var( dh, -1, IOCTL_GET_PCPS_DEV, p );
    return _mbgdevio_ret_val;
  #else
    *p = dh->dev;
    return PCPS_SUCCESS;
  #endif

}  // mbg_get_device_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_status_port( MBG_DEV_HANDLE dh, PCPS_STATUS_PORT *p )
{
  #if defined( _MBGIOCTL_H )
    _mbgdevio_vars();
    _mbgdevio_read_var( dh, -1, IOCTL_GET_PCPS_STATUS_PORT, p );
    return _mbgdevio_ret_val;
  #else
    *p = _pcps_ddev_read_status_port( dh );
    return PCPS_SUCCESS;
  #endif

}  // mbg_get_status_port



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_generic_read( MBG_DEV_HANDLE dh, int cmd,
                                             void *p, int size )
{
  _mbgdevio_vars();
  _mbgdevio_gen_read( dh, cmd, p, size );
  return _mbgdevio_ret_val;

}  // mbg_generic_read



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_generic_read_gps( MBG_DEV_HANDLE dh, int cmd,
                                                 void *p, int size )
{
  _mbgdevio_vars();
  _mbgdevio_gen_read_gps( dh, cmd, p, size );
  return _mbgdevio_ret_val;

}  // mbg_generic_read_gps



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_generic_write( MBG_DEV_HANDLE dh, int cmd,
                                              const void *p, int size )
{
  _mbgdevio_vars();
  _mbgdevio_gen_write( dh, cmd, p, size );
  return _mbgdevio_ret_val;

}  // mbg_generic_write



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_generic_write_gps( MBG_DEV_HANDLE dh, int cmd,
                                                  const void *p, int size )
{
  _mbgdevio_vars();
  _mbgdevio_gen_write_gps( dh, cmd, p, size );
  return _mbgdevio_ret_val;

}  // mbg_generic_write_gps



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_generic_io( MBG_DEV_HANDLE dh, int type,
                                           const void *in_p, int in_sz,
                                           void *out_p, int out_sz )
{
  _mbgdevio_vars();
  _mbgdevio_gen_io( dh, type, in_p, in_sz, out_p, out_sz );
  return _mbgdevio_ret_val;

}  // mbg_generic_io



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_time( MBG_DEV_HANDLE dh, PCPS_TIME *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var( dh, PCPS_GIVE_TIME, IOCTL_GET_PCPS_TIME, p );
  return _mbgdevio_ret_val;

}  // mbg_get_time



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_time( MBG_DEV_HANDLE dh, const PCPS_STIME *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_var_chk( dh, PCPS_SET_TIME, IOCTL_SET_PCPS_TIME, p,
                           _pcps_ddev_can_set_time( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_time



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_sync_time( MBG_DEV_HANDLE dh, PCPS_TIME *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GIVE_SYNC_TIME, IOCTL_GET_PCPS_SYNC_TIME,
                          p, _pcps_ddev_has_sync_time( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_sync_time



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_time_sec_change( MBG_DEV_HANDLE dh, PCPS_TIME *p )
{
  #if defined MBG_TGT_WIN32
    _mbgdevio_vars();
    _mbgdevio_read_var( dh, -1, IOCTL_GET_PCPS_TIME_SEC_CHANGE, p );
    return _mbgdevio_ret_val;
  #else
    #if defined( __BORLANDC__ )
      dh; p;  // avoid warnings "never used"
    #endif
    return -1;
  #endif

}  // mbg_get_time_sec_change



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_hr_time( MBG_DEV_HANDLE dh, PCPS_HR_TIME *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GIVE_HR_TIME, IOCTL_GET_PCPS_HR_TIME,
                          p, _pcps_ddev_has_hr_time( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_hr_time



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_event_time( MBG_DEV_HANDLE dh, const PCPS_TIME_STAMP *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_var_chk( dh, PCPS_SET_EVENT_TIME, IOCTL_SET_PCPS_EVENT_TIME,
                           p, _pcps_ddev_has_event_time( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_event_time



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_serial( MBG_DEV_HANDLE dh, PCPS_SERIAL *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var( dh, PCPS_GET_SERIAL, IOCTL_GET_PCPS_SERIAL, p );
  return _mbgdevio_ret_val;

}  // mbg_get_serial



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_serial( MBG_DEV_HANDLE dh, const PCPS_SERIAL *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_var( dh, PCPS_SET_SERIAL, IOCTL_SET_PCPS_SERIAL, p );
  return _mbgdevio_ret_val;

}  // mbg_set_serial



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_tzcode( MBG_DEV_HANDLE dh, PCPS_TZCODE *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GET_TZCODE, IOCTL_GET_PCPS_TZCODE,
                          p, _pcps_ddev_has_tzcode( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_tzcode



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_tzcode( MBG_DEV_HANDLE dh, const PCPS_TZCODE *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_var_chk( dh, PCPS_SET_TZCODE, IOCTL_SET_PCPS_TZCODE,
                           p, _pcps_ddev_has_tzcode( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_tzcode



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_pcps_tzdl( MBG_DEV_HANDLE dh, PCPS_TZDL *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GET_PCPS_TZDL, IOCTL_GET_PCPS_TZDL,
                          p, _pcps_ddev_has_pcps_tzdl( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_pcps_tzdl



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_pcps_tzdl( MBG_DEV_HANDLE dh, const PCPS_TZDL *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_var_chk( dh, PCPS_SET_PCPS_TZDL, IOCTL_SET_PCPS_TZDL,
                           p, _pcps_ddev_has_pcps_tzdl( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_pcps_tzdl



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_ref_offs( MBG_DEV_HANDLE dh, PCPS_REF_OFFS *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GET_REF_OFFS, IOCTL_GET_PCPS_REF_OFFS,
                          p, _pcps_ddev_has_ref_offs( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_ref_offs



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_ref_offs( MBG_DEV_HANDLE dh, const PCPS_REF_OFFS *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_var_chk( dh, PCPS_SET_REF_OFFS, IOCTL_SET_PCPS_REF_OFFS,
                           p, _pcps_ddev_has_ref_offs( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_ref_offs



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_opt_info( MBG_DEV_HANDLE dh, MBG_OPT_INFO *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GET_OPT_INFO, IOCTL_GET_MBG_OPT_INFO,
                          p, _pcps_ddev_has_opt_flags( dh ) );
 return _mbgdevio_ret_val;

}  // mbg_get_opt_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_opt_settings( MBG_DEV_HANDLE dh, const MBG_OPT_SETTINGS *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_var_chk( dh, PCPS_SET_OPT_SETTINGS,
                           IOCTL_SET_MBG_OPT_SETTINGS, p,
                           _pcps_ddev_has_opt_flags( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_opt_settings



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_irig_rx_info( MBG_DEV_HANDLE dh, IRIG_INFO *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GET_IRIG_RX_INFO, IOCTL_GET_PCPS_IRIG_RX_INFO,
                          p, _pcps_ddev_is_irig_rx( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_irig_rx_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_irig_rx_settings( MBG_DEV_HANDLE dh, const IRIG_SETTINGS *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_var_chk( dh, PCPS_SET_IRIG_RX_SETTINGS,
                           IOCTL_SET_PCPS_IRIG_RX_SETTINGS, p,
                           _pcps_ddev_is_irig_rx( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_irig_rx_settings



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_clr_ucap_buff( MBG_DEV_HANDLE dh )
{
  _mbgdevio_vars();
  _mbgdevio_write_cmd_chk( dh, PCPS_CLR_UCAP_BUFF, IOCTL_PCPS_CLR_UCAP_BUFF,
                           _pcps_ddev_can_clr_ucap_buff( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_clr_ucap_buff



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_ucap_entries( MBG_DEV_HANDLE dh, PCPS_UCAP_ENTRIES *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GIVE_UCAP_ENTRIES,
                          IOCTL_GET_PCPS_UCAP_ENTRIES, p,
                          _pcps_ddev_has_ucap( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_ucap_entries



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_ucap_event( MBG_DEV_HANDLE dh, PCPS_HR_TIME *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GIVE_UCAP_EVENT,
                          IOCTL_GET_PCPS_UCAP_EVENT, p,
                          _pcps_ddev_has_ucap( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_ucap_event



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_tzdl( MBG_DEV_HANDLE dh, TZDL *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_TZDL, IOCTL_GET_GPS_TZDL, p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_tzdl



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_tzdl( MBG_DEV_HANDLE dh, const TZDL *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_gps_var( dh, PC_GPS_TZDL, IOCTL_SET_GPS_TZDL, p );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_tzdl



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_sw_rev( MBG_DEV_HANDLE dh, SW_REV *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_SW_REV, IOCTL_GET_GPS_SW_REV, p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_sw_rev



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_bvar_stat( MBG_DEV_HANDLE dh, BVAR_STAT *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_BVAR_STAT, IOCTL_GET_GPS_BVAR_STAT, p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_stat



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_time( MBG_DEV_HANDLE dh, TTM *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_TIME, IOCTL_GET_GPS_TIME, p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_time



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_time( MBG_DEV_HANDLE dh, const TTM *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_gps_var( dh, PC_GPS_TIME, IOCTL_SET_GPS_TIME, p );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_time



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_port_parm( MBG_DEV_HANDLE dh, PORT_PARM *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_PORT_PARM, IOCTL_GET_GPS_PORT_PARM, p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_port_parm



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_port_parm( MBG_DEV_HANDLE dh, const PORT_PARM *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_gps_var( dh, PC_GPS_PORT_PARM, IOCTL_SET_GPS_PORT_PARM, p );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_port_parm



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_ant_info( MBG_DEV_HANDLE dh, ANT_INFO *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_ANT_INFO, IOCTL_GET_GPS_ANT_INFO, p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_ant_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_ucap( MBG_DEV_HANDLE dh, TTM *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_UCAP, IOCTL_GET_GPS_UCAP, p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_ucap



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_enable_flags( MBG_DEV_HANDLE dh, ENABLE_FLAGS *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_ENABLE_FLAGS,
                          IOCTL_GET_GPS_ENABLE_FLAGS, p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_enable_flags



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_enable_flags( MBG_DEV_HANDLE dh,
                                       const ENABLE_FLAGS *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_gps_var( dh, PC_GPS_ENABLE_FLAGS,
                           IOCTL_SET_GPS_ENABLE_FLAGS, p );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_enable_flags



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_stat_info( MBG_DEV_HANDLE dh, STAT_INFO *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_STAT_INFO, IOCTL_GET_GPS_STAT_INFO, p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_stat_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_cmd( MBG_DEV_HANDLE dh, const GPS_CMD *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_gps_var( dh, PC_GPS_CMD, IOCTL_SET_GPS_CMD, p );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_cmd


/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_pos( MBG_DEV_HANDLE dh, POS *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var( dh, PC_GPS_POS, IOCTL_GET_GPS_POS, p );
  swap_pos_doubles( p );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_pos



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_pos_xyz( MBG_DEV_HANDLE dh, const XYZ p )
{
  _mbgdevio_vars();
  XYZ xyz;
  int i;

  for ( i = 0; i < N_XYZ; i++ )
  {
    xyz[i] = p[i];
    swap_double( &xyz[i] );
  }

  _mbgdevio_write_gps_var( dh, PC_GPS_POS_XYZ, IOCTL_SET_GPS_POS_XYZ, xyz );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_pos_xyz



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_pos_lla( MBG_DEV_HANDLE dh, const LLA p )
{
  _mbgdevio_vars();
  LLA lla;
  int i;

  for ( i = 0; i < N_LLA; i++ )
  {
    lla[i] = p[i];
    swap_double( &lla[i] );
  }

  _mbgdevio_write_gps_var( dh, PC_GPS_POS_LLA, IOCTL_SET_GPS_POS_LLA, lla );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_pos_lla



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_ant_cable_len( MBG_DEV_HANDLE dh, ANT_CABLE_LEN *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var_chk( dh, PC_GPS_ANT_CABLE_LEN,
                              IOCTL_GET_GPS_ANT_CABLE_LEN, p,
                              _pcps_ddev_has_cab_len( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_ant_cable_len



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_ant_cable_len( MBG_DEV_HANDLE dh,
                                                      const ANT_CABLE_LEN *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_gps_var_chk( dh, PC_GPS_ANT_CABLE_LEN,
                               IOCTL_SET_GPS_ANT_CABLE_LEN, p,
                               _pcps_ddev_has_cab_len( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_ant_cable_len



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_receiver_info( MBG_DEV_HANDLE dh, RECEIVER_INFO *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_var_chk( dh, PC_GPS_RECEIVER_INFO,
                              IOCTL_GET_GPS_RECEIVER_INFO, p,
                              _pcps_ddev_has_receiver_info( dh ) );
         
  return _mbgdevio_ret_val;

}  // mbg_get_gps_receiver_info



static /*HDR*/  //##++
int _MBG_API mbg_get_gps_all_str_type_info( MBG_DEV_HANDLE dh,
                                            STR_TYPE_INFO_IDX stii[],
                                            const RECEIVER_INFO *p_ri )
{
  _mbgdevio_vars(); //##++
  _mbgdevio_read_gps_chk( dh, PC_GPS_ALL_STR_TYPE_INFO,
                          IOCTL_GET_GPS_ALL_STR_TYPE_INFO, stii,
                          p_ri->n_str_type * sizeof( stii[0] ),
                          _pcps_ddev_has_receiver_info( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_all_str_type_info



static /*HDR*/  //##++
int _MBG_API mbg_get_gps_all_port_info( MBG_DEV_HANDLE dh,
                                        PORT_INFO_IDX pii[],
                                        const RECEIVER_INFO *p_ri )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_chk( dh, PC_GPS_ALL_PORT_INFO,
                          IOCTL_GET_GPS_ALL_PORT_INFO, pii,
                          p_ri->n_com_ports * sizeof( pii[0] ),
                          _pcps_ddev_has_receiver_info( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_all_port_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_port_settings_idx( MBG_DEV_HANDLE dh,
                                                          const PORT_SETTINGS_IDX *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_gps_var_chk( dh, PC_GPS_PORT_SETTINGS_IDX,
                               IOCTL_SET_GPS_PORT_SETTINGS_IDX, p,
                               _pcps_ddev_has_receiver_info( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_port_settings_idx



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_port_settings( MBG_DEV_HANDLE dh,
                                                      const PORT_SETTINGS *p, int idx )
{
  PORT_SETTINGS_IDX psi = { 0 };

  psi.idx = idx;
  psi.port_settings = *p;

  return mbg_set_gps_port_settings_idx( dh, &psi );

}  // mbg_set_gps_port_settings



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_setup_receiver_info( MBG_DEV_HANDLE dh,
                                                    const PCPS_DEV *pdev,
                                                    RECEIVER_INFO *p )
{
  // If the clock supports the receiver_info structure then
  // read it from the clock, otherwise set up some default
  // values depending on the clock type.
  if ( _pcps_has_receiver_info( pdev ) )
    return mbg_get_gps_receiver_info( dh, p );


  if ( _pcps_is_gps( pdev ) )
  { // keep braces due to the macro call!
    _setup_default_receiver_info_gps( p );
  }
  else
  { // keep braces due to the macro call!
    _setup_default_receiver_info_dcf( p, pdev );
  }

  // Make sure this program supports at least as many ports as
  // the current clock device.
  if ( p->n_com_ports > MAX_PARM_PORT )
    return -1; //##++

  // Make sure this program supports at least as many string types
  // as the current clock device.
  if ( p->n_str_type > MAX_PARM_STR_TYPE )
    return -2; //##++


  return PCPS_SUCCESS;

}  // mbg_setup_receiver_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_serial_settings( MBG_DEV_HANDLE dh,
                                                    const PCPS_DEV *pdev,
                                                    RECEIVER_PORT_CFG *pcfg,
                                                    const RECEIVER_INFO *p_ri )
{
  int rc;
  int i;

  memset( pcfg, 0, sizeof( *pcfg ) );

  if ( _pcps_has_receiver_info( pdev ) )
  {
    rc = mbg_get_gps_all_port_info( dh, pcfg->pii, p_ri );
    if ( rc != PCPS_SUCCESS )
      goto error;

    rc = mbg_get_gps_all_str_type_info( dh, pcfg->stii, p_ri );
    if ( rc != PCPS_SUCCESS )
      goto error;
  }
  else
  {
    if ( _pcps_is_gps( pdev ) )
    {
      rc = mbg_get_gps_port_parm( dh, &pcfg->tmp_pp );
      if ( rc != PCPS_SUCCESS )
        goto error;

      for ( i = 0; i < p_ri->n_com_ports; i++ )
      {
        PORT_INFO_IDX *p_pii = &pcfg->pii[i];
        PORT_INFO *p_pi = &p_pii->port_info;

        p_pii->idx = i;
        port_settings_from_port_parm( &p_pi->port_settings,
                                      i, &pcfg->tmp_pp, 1 );

        p_pi->supp_baud_rates = DEFAULT_GPS_BAUD_RATES_C166;
        p_pi->supp_framings = DEFAULT_GPS_FRAMINGS_C166;
        p_pi->supp_str_types = DEFAULT_SUPP_STR_TYPES_GPS;
      }
    }
    else
    {
      PCPS_SERIAL ser_code;

      rc = mbg_get_serial( dh, &ser_code );
      if ( rc != PCPS_SUCCESS )
        goto error;


      port_info_from_pcps_serial( pcfg->pii, ser_code,
                                  _pcps_has_serial_hs( pdev ) ?
                                    DEFAULT_BAUD_RATES_DCF_HS :
                                    DEFAULT_BAUD_RATES_DCF
                               );
    }

    for ( i = 0; i < p_ri->n_str_type; i++ )
    {
      STR_TYPE_INFO_IDX *stip = &pcfg->stii[i];
      stip->idx = i;
      stip->str_type_info = default_str_type_info[i];
    }
  }

  return PCPS_SUCCESS;


error:
  return rc;

}  // mbg_get_serial_settings



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_save_serial_settings( MBG_DEV_HANDLE dh,
                                                     const PCPS_DEV *pdev,
                                                     RECEIVER_PORT_CFG *pcfg,
                                                     int port_num )
{
  int rc;

  if ( _pcps_has_receiver_info( pdev ) )
  {
    rc = mbg_set_gps_port_settings( dh, &pcfg->pii[port_num].port_info.port_settings, port_num );
  }
  else
  {
    if ( _pcps_is_gps( pdev ) )
    {
      port_parm_from_port_settings( &pcfg->tmp_pp, port_num,
                          &pcfg->pii[port_num].port_info.port_settings, 1 );

      rc = mbg_set_gps_port_parm( dh, &pcfg->tmp_pp );
    }
    else
    {
      PCPS_SERIAL ser_code;

      pcps_serial_from_port_info( &ser_code, pcfg->pii );

      rc = mbg_set_serial( dh, &ser_code );
    }
  }

  return rc;

}  // mbg_save_serial_settings



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_asic_version( MBG_DEV_HANDLE dh, PCPS_ASIC_VERSION *p )
{

  #if defined( _MBGIOCTL_H )
    _mbgdevio_vars();
    _mbgdevio_read_var( dh, -1, IOCTL_GET_PCPS_ASIC_VERSION, p );
    return _mbgdevio_ret_val;
  #else
    if ( !_pcps_ddev_is_pci_asic( dh ) )
      return -1;

    *p = inpd( _pcps_ddev_port_base( dh, 0 )
             + offsetof( PCI_ASIC, features ) );

    return PCPS_SUCCESS;
  #endif

}  // mbg_get_asic_version



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_time_cycles( MBG_DEV_HANDLE dh, PCPS_TIME_CYCLES *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var( dh, PCPS_GIVE_TIME, IOCTL_GET_PCPS_TIME_CYCLES, p );
  #if !defined( _MBGIOCTL_H )
    // only if not using IOCTLs
    // for PCPS_TIME, read stamp AFTER the call
    p->cycles = 0;  //##++
  #endif
  return _mbgdevio_ret_val;

}  // mbg_get_time_cycles



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_hr_time_cycles( MBG_DEV_HANDLE dh,
                                                   PCPS_HR_TIME_CYCLES *p )
{
  _mbgdevio_vars();
  #if !defined( _MBGIOCTL_H )
    // only if not using IOCTLs
    // for PCPS_TIME, read stamp BEFORE the call
    p->cycles = 0;  //##++
  #endif
  _mbgdevio_read_var_chk( dh, PCPS_GIVE_HR_TIME,
                          IOCTL_GET_PCPS_HR_TIME_CYCLES,
                          p, _pcps_ddev_has_hr_time( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_hr_time_cycles



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_hr_time_comp( MBG_DEV_HANDLE dh, PCPS_HR_TIME *p,
                                                 int32_t *hns_latency )
{
  #if defined( MBG_TGT_WIN32 )

    PCPS_HR_TIME_CYCLES htc;
    LARGE_INTEGER pc;
    int rc;

    // First get current performance counter value, then read
    // a high resolution time stamp from the board, plus the
    // associated performance counter value.
    QueryPerformanceCounter( &pc );
    rc = mbg_get_hr_time_cycles( dh, &htc );

    if ( rc == PCPS_SUCCESS )
    {
      static LARGE_INTEGER performanceFrequency;

      LONGLONG raw_latency;
      LONGLONG frac_latency;
      LARGE_INTEGER comp_frac;

      // Performance counter frequency must only be read once.
      if ( performanceFrequency.QuadPart == 0 )
        QueryPerformanceFrequency( &performanceFrequency ) ;

      // Compute the latency in performance counter units.
      raw_latency = htc.cycles.QuadPart - pc.QuadPart;

      // Account for performance counter overflow. This is
      // supposed to happen once every 2^^63 units of the
      // performance counter frequency, i.e. about every
      // 97 years on a system with 3 GHz clock.
      if ( raw_latency < 0 )
        raw_latency += MAXLONGLONG;

      // convert latency to binary fractions of seconds,
      // i.e. units of 2^^-32.
      frac_latency = raw_latency * ( ( (int64_t) 1 ) << 32 )
                     / performanceFrequency.QuadPart;

      // compute the compensated fractional part of the HR time stamp
      // and account for borrows from the sec field
      comp_frac.QuadPart = htc.t.tstamp.frac - frac_latency;
      htc.t.tstamp.frac = comp_frac.LowPart;
      htc.t.tstamp.sec -= comp_frac.HighPart;

      *p = htc.t;

      if ( hns_latency )
      {
        LONGLONG tmp_hns_latency;

        // convert to hectonanoseconds
        tmp_hns_latency = raw_latency * 10000000 / performanceFrequency.QuadPart;

        // check for range overflow
        if ( tmp_hns_latency > MAXLONG )
          tmp_hns_latency = MAXLONG;
        else
          if ( tmp_hns_latency < ( -MAXLONG - 1 ) )  // don't use MINLONG here
            tmp_hns_latency = -MAXLONG - 1;          // due to sign extension probs

        *hns_latency = (int32_t) tmp_hns_latency;
      }
    }

    return rc;

  #else

    return -1;   // not yet supported

  #endif

}  // mbg_get_hr_time_comp



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_irig_tx_info( MBG_DEV_HANDLE dh, IRIG_INFO *p )
{
  _mbgdevio_vars();

  #if !defined( _MBGIOCTL_H )
    // This is a workaround for GPS169PCIs with early
    // firmware versions. See RCS log for details.
    uint8_t pcps_cmd = PCPS_GET_IRIG_TX_INFO;

    if ( _pcps_ddev_requires_irig_workaround( dh ) )
      pcps_cmd = PCPS_GET_IRIG_RX_INFO;

    #define _PCPS_CMD   pcps_cmd
  #else
    #define _PCPS_CMD   PCPS_GET_IRIG_TX_INFO
  #endif

  _mbgdevio_read_var_chk( dh, _PCPS_CMD, IOCTL_GET_PCPS_IRIG_TX_INFO,
                          p, _pcps_ddev_has_irig_tx( dh ) );
  #undef _PCPS_CMD

  return _mbgdevio_ret_val;

}  // mbg_get_irig_tx_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_irig_tx_settings( MBG_DEV_HANDLE dh, const IRIG_SETTINGS *p )
{
  _mbgdevio_vars();

  #if !defined( _MBGIOCTL_H )
    // This is a workaround for GPS169PCIs with early
    // firmware versions. See RCS log for details.
    uint8_t pcps_cmd = PCPS_SET_IRIG_TX_SETTINGS;

    if ( _pcps_ddev_requires_irig_workaround( dh ) )
      pcps_cmd = PCPS_SET_IRIG_RX_SETTINGS;

    #define _PCPS_CMD   pcps_cmd
  #else
    #define _PCPS_CMD   PCPS_SET_IRIG_TX_SETTINGS
  #endif

  _mbgdevio_write_var_chk( dh, _PCPS_CMD, IOCTL_SET_PCPS_IRIG_TX_SETTINGS,
                           p, _pcps_ddev_has_irig_tx( dh ) );

  #undef _PCPS_CMD

  return _mbgdevio_ret_val;

}  // mbg_set_irig_tx_settings



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_synth( MBG_DEV_HANDLE dh, SYNTH *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GET_SYNTH, IOCTL_GET_SYNTH,
                          p, _pcps_ddev_has_synth( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_synth



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_synth( MBG_DEV_HANDLE dh, const SYNTH *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_var_chk( dh, PCPS_SET_SYNTH, IOCTL_SET_SYNTH, 
                           p, _pcps_ddev_has_synth( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_synth



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_synth_state( MBG_DEV_HANDLE dh, SYNTH_STATE *p )
{
  _mbgdevio_vars();
  _mbgdevio_read_var_chk( dh, PCPS_GET_SYNTH_STATE, IOCTL_GET_SYNTH_STATE,
                          p, _pcps_ddev_has_synth( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_synth_state



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_is_gps( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_is_gps, IOCTL_DEV_IS_GPS, p );

}  // mbg_dev_is_gps



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_is_dcf( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_is_dcf, IOCTL_DEV_IS_DCF, p );

}  // mbg_dev_is_dcf



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_is_irig_rx( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_is_irig_rx, IOCTL_DEV_IS_IRIG_RX, p );

}  // mbg_dev_is_irig_rx



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_hr_time( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_hr_time, IOCTL_DEV_HAS_HR_TIME, p );

}  // mbg_dev_has_hr_time



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_cab_len( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_cab_len, IOCTL_DEV_HAS_CAB_LEN, p );

}  // mbg_dev_has_cab_len



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_tzdl( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_tzdl, IOCTL_DEV_HAS_TZDL, p );

}  // mbg_dev_has_tzdl



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_pcps_tzdl( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_pcps_tzdl, IOCTL_DEV_HAS_PCPS_TZDL, p );

}  // mbg_dev_has_pcps_tzdl



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_tzcode( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_tzcode, IOCTL_DEV_HAS_TZCODE, p );

}  // mbg_dev_has_tzcode



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_tz( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_tz, IOCTL_DEV_HAS_TZ, p );

}  // mbg_dev_has_tz



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_event_time( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_event_time, IOCTL_DEV_HAS_EVENT_TIME, p );

}  // mbg_dev_has_event_time



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_receiver_info( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_receiver_info, IOCTL_DEV_HAS_RECEIVER_INFO, p );

}  // mbg_dev_has_receiver_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_can_clr_ucap_buff( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_can_clr_ucap_buff, IOCTL_DEV_CAN_CLR_UCAP_BUFF, p );

}  // mbg_dev_can_clr_ucap_buff



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_ucap( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_ucap, IOCTL_DEV_HAS_UCAP, p );

}  // mbg_dev_has_ucap



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_irig_tx( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_irig_tx, IOCTL_DEV_HAS_IRIG_TX, p );

}  // mbg_dev_has_irig_tx



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_serial_hs( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_serial_hs, IOCTL_DEV_HAS_SERIAL_HS, p );

}  // mbg_dev_has_serial_hs



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_signal( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_signal, IOCTL_DEV_HAS_SIGNAL, p );

}  // mbg_dev_has_signal



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_mod( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_mod, IOCTL_DEV_HAS_MOD, p );

}  // mbg_dev_has_mod



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_irig( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_irig, IOCTL_DEV_HAS_IRIG, p );

}  // mbg_dev_has_irig



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_ref_offs( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_ref_offs, IOCTL_DEV_HAS_REF_OFFS, p );

}  // mbg_dev_has_ref_offs



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_opt_flags( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_opt_flags, IOCTL_DEV_HAS_OPT_FLAGS, p );

}  // mbg_dev_has_opt_flags



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_gps_data( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_gps_data, IOCTL_DEV_HAS_GPS_DATA, p );

}  // mbg_dev_has_gps_data



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_synth( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_synth, IOCTL_DEV_HAS_SYNTH, p );

}  // mbg_dev_has_synth



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_dev_has_generic_io( MBG_DEV_HANDLE dh, int *p )
{
  _mbgdevio_query_cond( dh, _pcps_ddev_has_generic_io, IOCTL_DEV_HAS_GENERIC_IO, p );

}  // mbg_dev_has_generic_io



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_get_gps_all_pout_info( MBG_DEV_HANDLE dh,
                                        POUT_INFO_IDX pii[],
                                        const RECEIVER_INFO *p_ri )
{
  _mbgdevio_vars();
  _mbgdevio_read_gps_chk( dh, PC_GPS_ALL_POUT_INFO,
                          IOCTL_GET_GPS_ALL_POUT_INFO, pii,
                          p_ri->n_prg_out * sizeof( pii[0] ),
                          _pcps_ddev_has_receiver_info( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_get_gps_all_pout_info



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_pout_settings_idx( MBG_DEV_HANDLE dh,
                                                          const POUT_SETTINGS_IDX *p )
{
  _mbgdevio_vars();
  _mbgdevio_write_gps_var_chk( dh, PC_GPS_POUT_SETTINGS_IDX,
                               IOCTL_SET_GPS_POUT_SETTINGS_IDX, p,
                               _pcps_ddev_has_receiver_info( dh ) );
  return _mbgdevio_ret_val;

}  // mbg_set_gps_pout_settings_idx



/*HDR*/
_MBG_API_ATTR int _MBG_API mbg_set_gps_pout_settings( MBG_DEV_HANDLE dh,
                                                      const POUT_SETTINGS *p, int idx )
{
  POUT_SETTINGS_IDX psi = { 0 };

  psi.idx = idx;
  psi.pout_settings = *p;

  return mbg_set_gps_pout_settings_idx( dh, &psi );

}  // mbg_set_gps_pout_settings



#if defined( MBG_TGT_WIN32 )

#if defined( _USRDLL )

BOOL APIENTRY DllMain( HANDLE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved )
{
  return TRUE;
}

#endif // defined( _USRDLL )

#endif  // defined( MBG_TGT_WIN32 )

