
/**************************************************************************
 *
 *  $Id: pcpsdrvr.c 1.28 2006/07/11 10:24:20 martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Driver functions that detect Meinberg PC plug-in devices and set up
 *    the software environment (port base address, clock features, etc.).
 *
 *    These functions should be used with programs which have direct
 *    access to the hardware (e.g. device drivers).
 *
 *    Programs which access the devices via device drivers should
 *    use the functions provided by the mbgdevio module.
 *
 *    There are several preprocessor symbols defined at the top of
 *    pcpsdrvr.h which control the default support of some features
 *    under the different operating systems. if required, each of
 *    those symbols can be overridden by compiler arguments.
 *
 *    Basically the following radio clocks are supported:
 *      PCI bus 5V:      PCI32, GPS167PCI, PCI509, GPS168PCI
 *      PCI bus 5V/3.3V: PCI510, PCI511, GPS169PCI, GPS170PCI,
 *                       TCR510PCI, TCR167PCI, TCR511PCI
 *      MCA bus:         PS31
 *      ISA bus:         PC31, PC32, GPS167PC
 *
 *    PCI support is possible in two different ways. The preferred
 *    functions are compiled in if one of the symbols _PCPS_USE_PCIMGR
 *    or _PCPS_USE_PCIBIOS is defined != 0.
 *
 *    If _PCPS_USE_PCIMGR is != 0 it is assumed that the operating
 *    system's PCI layer detects a new PCI device and calls a driver's
 *    add_device()/start_device() function to initialize the device.
 *    This new technique is supported with PNP operating systems
 *    (e.g. Win98, Win2K, newer Linux versions).
 *
 *    If _PCPS_USE_PCIBIOS is != 0 the program scans the PCI bus
 *    during startup to detect and initialize supported PCI devices.
 *    This techique is used with non-PNP operating systems.
 *
 *    The symbol _PCPS_USE_RSRCMGR must be defined != 0 to include
 *    support of resource managers, if necessary.
 *
 *    If the symbol _PCPS_USE_MCA is defined != 0 then Micro Channel
 *    detection (and therefore auto-detection of a MCA clock) is
 *    supported.
 *
 *    MCA clocks are accessed using the same low level functions as
 *    ISA clocks, so if autodetection of MCA clocks is not supported
 *    then a MCA clock's known port number can be passed to
 *    pcps_detect_clocks() to let it be treated like an ISA clock.
 *
 * -----------------------------------------------------------------------
 *  $Log: pcpsdrvr.c $
 *  Revision 1.28  2006/07/11 10:24:20  martin
 *  Use _fmemcpy() in pcps_generic_io() to support environments which
 *  require far data pointers.
 *  Revision 1.27  2006/07/07 09:41:15  martin
 *  Renamed pci_..() function calls to _mbg_pci_..() calls which are defined according to the
 *  OS requirements, in order to avoid naming conflicts.
 *  Revision 1.26  2006/06/19 15:28:52  martin
 *  Added support for TCR511PCI.
 *  Modified parameters required to detect ISA cards. 
 *  The array of port addresses does no more require a 0 address
 *  as last value.
 *  Revision 1.25  2006/03/10 11:01:27  martin
 *  Added support for PCI511.
 *  Revision 1.24  2005/11/03 15:50:45Z  martin
 *  Added support for GPS170PCI.
 *  Revision 1.23  2005/09/16 08:21:08Z  martin
 *  Also flag PCI cards which have base_addr set to 0 as uninitialized.
 *  Revision 1.22  2005/06/02 10:32:07Z  martin
 *  Changed more types to C99 fixed size types.
 *  New function pcps_generic_io().
 *  Revision 1.21  2004/12/13 14:19:38Z  martin
 *  Support configuration of on-board frequency synthesizer.
 *  Revision 1.20  2004/11/09 13:02:48Z  martin
 *  Redefined fixed width data types using standard C99 types.
 *  Fixed warnings about lvalue casts.
 *  Revision 1.19  2004/10/14 15:01:24  martin
 *  Added support for TCR167PCI.
 *  Revision 1.18  2004/09/06 15:16:57Z  martin
 *  Support a GPS_DATA interface where sizes are specified 
 *  by 16 instead of the original 8 bit quantities, thus allowing 
 *  to transfer data blocks which exceed 255 bytes.
 *  Conditionally skip assertions under Linux.
 *  Revision 1.17  2004/04/22 14:47:54  martin
 *  Fixed conversion of firmware rev. number.
 *  Revision 1.16  2004/04/07 09:45:04Z  martin
 *  Support new feature PCPS_HAS_IRIG_TX for GPS169PCI.
 *  Revision 1.15  2003/12/22 16:15:21Z  martin
 *  Support PCPS_HR_TIME for TCR510PCI.
 *  Revision 1.14  2003/07/30 07:28:23Z  martin
 *  Moved prototype for register_pci_devices() outside to top of file.
 *  Revision 1.13  2003/07/08 15:11:55  martin
 *  Support PCI PNP interface under Linux.
 *  New function pcps_rsrc_release().
 *  Made some functions public.
 *  Renamed some public functions to start with pcps_...
 *  Revision 1.12  2003/06/19 10:08:43  MARTIN
 *  Renamed some functions to follow common naming conventions.
 *  Made a function's parameter pointer const.
 *  Changes due to renamed symbols in pcpsdev.h.
 *  Check devices for _pcps_has_ucap() support.
 *  Revision 1.11  2003/05/16 09:28:06  MARTIN
 *  Moved inclusion of some headers to pcpsdrvr.h.
 *  Revision 1.10  2003/04/09 16:35:57  martin
 *  Supports PCI510, GPS169PCI, and TCR510PCI,
 *  and new PCI_ASIC used by those devices.
 *  Revision 1.9  2003/03/20 11:42:37  martin
 *  Fixed syntax for QNX.
 *  Revision 1.8  2002/08/09 08:25:50  MARTIN
 *  Support feature PCPS_CAN_CLR_CAP_BUFF.
 *  Fixed a bug resulting in an unterminated string
 *  if SERNUM was being read.
 *  Revision 1.7  2002/02/26 09:31:57  MARTIN
 *  New function pcps_read_sernum().
 *  Revision 1.6  2002/02/19 09:46:26  MARTIN
 *  Use new header mbg_tgt.h to check the target environment.
 *  Removed function pcps_sn_str_from_ident(), use new
 *  function mbg_gps_ident_decode() from identdec.c now.
 *  If a PCI clock's interface is not properly configured don't 
 *  enable the device and set the read function to the new 
 *  dummy function pcps_read_null() to prevent driver from 
 *  accessing random ports.
 *  Revision 1.5  2002/02/01 12:06:12  MARTIN
 *  Added support for GPS168PCI.
 *  Removed obsolete code.
 *  Revision 1.4  2001/11/30 09:52:48  martin
 *  Added support for event_time which, however, requires
 *  a custom GPS firmware.
 *  Revision 1.3  2001/09/18 06:59:18  MARTIN
 *  Account for new preprocessor symbols in the header file.
 *  Added some type casts to avoid compiler warnings under Win32.
 *  Added some debug messages to clock detection functions.
 *  Revision 1.2  2001/03/16 14:45:33  MARTIN
 *  New functions and definitions to support PNP drivers.
 *  Revision 1.1  2001/03/01 16:26:41  MARTIN
 *  Initial revision for the new library.
 *
 **************************************************************************/

#define _PCPSDRVR
  #include <pcpsdrvr.h>
#undef _PCPSDRVR

#include <identdec.h>
#include <mbgddmsg.h>
#include <pci_asic.h>
#include <amccdefs.h>

#if !defined( MBG_TGT_LINUX )
  #include <assert.h>
#endif

#if _PCPS_USE_MCA
  #include <mca.h>
#endif

#if _PCPS_USE_PCI
  #include <pci.h>
#endif


// In some environments special far functions are are neither
// required nor supported, so redefine calls to those functions
// to appropriate standard function calls.
#if defined( MBG_TGT_NETWARE ) || defined( MBG_TGT_WIN32 ) || \
    defined( MBG_TGT_LINUX ) || defined( MBG_TGT_QNX )
  #define _fmemcpy( _d, _s, _n )      memcpy( _d, _s, _n )
  #define _fstrlen( _s )              strlen( _s )
  #define _fstrncmp( _s1, _s2, _n )   strncmp( (_s1), (_s2), (_n) )
#elif defined( MBG_TGT_OS2 )
  #define _fstrncmp( _s1, _s2, _n )   _fmemcmp( (_s1), (_s2), (_n) )
#endif

#if defined( MBG_TGT_OS2 )
  // Watcom C Compiler options for the OS/2 device driver result in
  // warnings if automatic stack addresses are passed to functions.
  #define static_wc static
  #define FMT_03X "%X"
  #define FMT_08X "%X"
#else
  #define static_wc
  #define FMT_03X "%03X"
  #define FMT_08X "%08lX"
#endif

#if defined( MBG_TGT_LINUX )
  typedef unsigned int PCI_DWORD;

  #if _PCPS_USE_PCIMGR
    // The function below is called to let the kernel know
    // which PCI clocks are supported by the driver.
    extern int register_pci_devices( void );
  #endif

#else
  typedef dword PCI_DWORD;
#endif

#if _PCPS_USE_RSRCMGR
  #if defined( MBG_TGT_OS2 )
    #define _RSRC_DEV       /* register devices */
    #define _RSRC_DRVR      /* register driver */
  #endif
#endif

#ifdef _RSRC_DEV
  #define RSRC_BASE_NAME "RADIOCLK_#  Meinberg Radio Clock "
  static const char *rsrc_clock_dcf77 = RSRC_BASE_NAME "(DCF77)";
  static const char *rsrc_clock_gps = RSRC_BASE_NAME "(GPS)";
#endif



/*--------------------------------------------------------------
 * Name:         pcps_read_null()
 *               pcps_read_std()
 *               pcps_read_amcc_s5933()
 *               pcps_read_amcc_s5920()
 *               pcps_read_asic()
 *
 * Purpose:      These functions are used for low level access
 *               to Meinberg plug-in radio clocks. The function
 *               to be used depends on the clock's bus type and
 *               interface chip.
 *
 * Input:        pcfg    pointer to the clock's configuration
 *               cmd     the command code for the board
 *               count   the number of bytes to be read
 *
 * Output:       buffer    the bytes that could be read
 *
 * Ret value:    PCPS_SUCCESS      no error
 *               PCPS_ERR_TIMEOUT  board is busy for too long
 *-------------------------------------------------------------*/

// The dummy read function below is used if a clock is
// not properly initialized, in order to avoid I/O access
// on unspecified ports.

static /*HDR*/   /* type: PCPS_READ_FNC */
short pcps_read_null( PCPS_DEV_CFG *pcfg, uint8_t cmd,
                      uint8_t FAR *buffer, uint8_t count )
{

  return PCPS_ERR_TIMEOUT;

}  // pcps_read_null



// The function below must be used to access a clock with
// standard ISA or micro channel bus.

static /*HDR*/   /* type: PCPS_READ_FNC */
short pcps_read_std( PCPS_DEV_CFG *pcfg, uint8_t cmd,
                     uint8_t FAR *buffer, uint8_t count )
{
  uint16_t port = _pcps_cfg_port_base( pcfg, 0 );
  clock_t timer;
  int i;


  // write the command byte
  outp( port, cmd );

  // wait until BUSY flag goes low or timeout
  #if _PCPS_USE_CLOCK_TICK
    timer = clock() + pcfg->timeout_clk;

    while ( _pcps_cfg_read_status_port( pcfg ) & PCPS_ST_BUSY )
      if ( clock() >= timer )
        return PCPS_ERR_TIMEOUT;
  #else
    for ( timer = pcfg->timeout_clk;
          _pcps_cfg_read_status_port( pcfg ) & PCPS_ST_BUSY; timer-- )
      if ( timer == 0 )
        return PCPS_ERR_TIMEOUT;
  #endif


  // no timeout: read bytes from the board's FIFO
  for ( i = 0; i < count; i++ )
    *buffer++ = (uint8_t) inp( port );

  return PCPS_SUCCESS;

} // pcps_read_std



#if _PCPS_USE_PCI

// The function below must be used to access a clock with
// PCI bus and AMCC S5933 interface chip.

static /*HDR*/   /* type: PCPS_READ_FNC */
short pcps_read_amcc_s5933( PCPS_DEV_CFG *pcfg, uint8_t cmd,
                            uint8_t FAR *buffer, uint8_t count )
{
  uint16_t port = _pcps_cfg_port_base( pcfg, 0 );
  clock_t timer;
  int i;


  // reset inbound mailbox and FIFO status
  outp( (uint16_t) ( port + AMCC_OP_REG_MCSR + 3 ), 0x0C );

  // set FIFO
  outp( (uint16_t) ( port + AMCC_OP_REG_INTCSR + 3 ), 0x3C );

  // write the command byte
  outp( (uint16_t) ( port + AMCC_OP_REG_OMB1 ), cmd );

  // wait until BUSY flag goes low or timeout
  #if _PCPS_USE_CLOCK_TICK
    timer = clock() + pcfg->timeout_clk;

    while ( _pcps_cfg_read_status_port( pcfg ) & PCPS_ST_BUSY )
      if ( clock() >= timer )
        return PCPS_ERR_TIMEOUT;
  #else
    for ( timer = pcfg->timeout_clk;
          _pcps_cfg_read_status_port( pcfg ) & PCPS_ST_BUSY; timer-- )
      if ( timer == 0 )
        return PCPS_ERR_TIMEOUT;
  #endif


  // no timeout: read bytes from the board's FIFO
  for ( i = 0; i < count; i++ )
  {
    if ( inpw( (uint16_t) ( port + AMCC_OP_REG_MCSR ) ) & 0x20 )
      return PCPS_ERR_FIFO;

    buffer[i] = (uint8_t) inp( (uint16_t) ( port + AMCC_OP_REG_FIFO + ( i % 4 ) ) );
  }

  return PCPS_SUCCESS;

}  /* pcps_read_amcc_s5933 */

#endif  /* _PCPS_USE_PCI */



#if _PCPS_USE_PCI

// The function below must be used to access a clock with
// PCI bus and AMCC S5920 interface chip.

static /*HDR*/   /* type: PCPS_READ_FNC */
short pcps_read_amcc_s5920( PCPS_DEV_CFG *pcfg, uint8_t cmd,
                            uint8_t FAR *buffer, uint8_t count )
{
  uint8_t FAR *p = buffer;
  uint16_t data_port = _pcps_cfg_port_base( pcfg, 1 );
  int i;

  clock_t timer;
  int dt_quot;
  int dt_rem;


  // write the command byte
  outp( (uint16_t) ( _pcps_cfg_port_base( pcfg, 0 ) + AMCC_OP_REG_OMB ), cmd );

  dt_quot = count / 4;
  dt_rem = count % 4;

  #if _PCPS_USE_CLOCK_TICK
    timer = clock() + pcfg->timeout_clk;

    while ( _pcps_cfg_read_status_port( pcfg ) & PCPS_ST_BUSY )
      if ( clock() >= timer )
        return PCPS_ERR_TIMEOUT;
  #else
    for ( timer = pcfg->timeout_clk;
          _pcps_cfg_read_status_port( pcfg ) & PCPS_ST_BUSY; timer-- )
      if ( timer == 0 )
        return PCPS_ERR_TIMEOUT;
  #endif


  if ( count )
  {
    // do this only if we must read data

    // first read full 32 bit words
    for ( i = 0; i < dt_quot; i++ )
    {
      * (uint32_t FAR *) p = inpd( data_port );
      p += sizeof( uint32_t );
    }

    // then read the remaining bytes, if required
    if ( dt_rem )
    {
      #define BYTE_OF( _x, _n )  *( ( (uint8_t *) &(_x) ) + (_n) )
      uint32_t ul = inpd( data_port );

      for ( i = 0; i < dt_rem; i++ )
        *p++ = BYTE_OF( ul, i );
    }
  }
  else
    inpd( data_port );  // do a dummy read

  return PCPS_SUCCESS;

}  // pcps_read_amcc_s5920

#endif  /* _PCPS_USE_PCI */



#if _PCPS_USE_PCI

// The function below must be used to access a clock with
// PCI bus and Meinberg PCI interface ASIC.

static /*HDR*/   /* type: PCPS_READ_FNC */
short pcps_read_asic( PCPS_DEV_CFG *pcfg, uint8_t cmd,
                      uint8_t FAR *buffer, uint8_t count )
{
  uint8_t FAR *p = buffer;
  uint16_t data_port;
  int i;

  clock_t timer;
  int dt_quot;
  int dt_rem;


  // write the command byte
  outpd( (uint16_t) ( _pcps_cfg_port_base( pcfg, 0 )
                      + offsetof( PCI_ASIC, pci_data ) ), cmd );

  data_port = _pcps_cfg_port_base( pcfg, 0 )
                + offsetof( PCI_ASIC, addon_data );
  dt_quot = count / 4;
  dt_rem = count % 4;

  #if _PCPS_USE_CLOCK_TICK
    timer = clock() + pcfg->timeout_clk;

    while ( _pcps_cfg_read_status_port( pcfg ) & PCPS_ST_BUSY )
      if ( clock() >= timer )
        return PCPS_ERR_TIMEOUT;
  #else
    for ( timer = pcfg->timeout_clk;
          _pcps_cfg_read_status_port( pcfg ) & PCPS_ST_BUSY; timer-- )
      if ( timer == 0 )
        return PCPS_ERR_TIMEOUT;
  #endif


  // no timeout: read bytes from the board's FIFO

  // first read full 32 bit words
  for ( i = 0; i < dt_quot; i++ )
  {
    uint32_t ul = inpd( data_port );
    * (uint32_t FAR *) p = ul;
    p += sizeof( uint32_t );
    data_port += sizeof( uint32_t );
  }

  // then read the remaining bytes, if required
  if ( dt_rem )
  {
    PCI_ASIC_REG ar;
    ar.ul = inpd( data_port );

    for ( i = 0; i < dt_rem; i++ )
      *p++ = ar.b[i];
  }

  return PCPS_SUCCESS;

}  // pcps_read_asic

#endif  /* _PCPS_USE_PCI */



/*--------------------------------------------------------------
 * Name:         pcps_write()
 *
 * Purpose:      Write data to the radio clock.
 *
 * Input:        pddev       pointer to the device information
 *               cmd         the address of buffer holding the
 *                           date/time/status information
 *               read_fnc    function to access the board
 *
 * Output:       --
 *
 * Ret value:    PCPS_SUCCESS      no error
 *               PCPS_ERR_TIMEOUT  board is busy for too long
 *               PCPS_ERR_NBYTES   the number of parameter bytes
 *                                 did not match the number of
 *                                 data bytes expected
 *               PCPS_ERR_STIME    the date, time or status
 *                                 has been invalid
 *-------------------------------------------------------------*/

/*HDR*/
short pcps_write( PCPS_DDEV *pddev, uint8_t cmd,
                  const void FAR *buffer, uint8_t count )
{
  const uint8_t FAR *p = (uint8_t FAR *) buffer;
  int i;
  short rc;
  uint8_t bytes_expected;
  int8_t write_rc;


  // Write the command and read one byte which will contain
  // the number of data bytes that must follow.
  rc = _pcps_read_var( pddev, cmd, bytes_expected );

  if ( rc < 0 )
    return rc;


  // Check if the number of data bytes to be written is correct.
  if ( bytes_expected != count )
    return PCPS_ERR_NBYTES;


  // Write all bytes but the last one without reading anything.
  bytes_expected--;

  for ( i = 0; i < bytes_expected; i++ )
  {
    rc = _pcps_write_byte( pddev, *p++ );

    if ( rc < 0 )
      return rc;
  }

  // Write the last byte and read the completion code.
  rc = _pcps_read_var( pddev, *p++, write_rc );

  // If an error code has been returned by the I/O function,
  // return that code, otherwise return the completion code
  // read from the board.
  return ( rc < 0 ) ? rc : write_rc;

}  // pcps_write



/*--------------------------------------------------------------
 * Name:         pcps_generic_io()
 *
 * Purpose:      Write data to the radio clock.
 *
 * Input:        pddev       pointer to the device information
 *               cmd         the address of buffer holding the
 *                           date/time/status information
 *               read_fnc    function to access the board
 *
 * Output:       --
 *
 * Ret value:    PCPS_SUCCESS      no error
 *               PCPS_ERR_TIMEOUT  board is busy for too long
 *               PCPS_ERR_NBYTES   the number of parameter bytes
 *                                 did not match the number of
 *                                 data bytes expected
 *               PCPS_ERR_STIME    the date, time or status
 *                                 has been invalid
 *-------------------------------------------------------------*/

/*HDR*/
short pcps_generic_io( PCPS_DDEV *pddev, uint8_t type,
                       const void FAR *in_buff, uint8_t in_cnt,
                       void FAR *out_buff, uint8_t out_cnt )
{
  const uint8_t FAR *p;
  int i;
  short rc;
  uint8_t tmp_byte;
  int8_t data_read[PCPS_FIFO_SIZE];
  int bytes_to_read;

  // Write the command and read one byte which will contain
  // the number of data bytes that must follow.
  rc = _pcps_read_var( pddev, PCPS_GENERIC_IO, tmp_byte );

  if ( rc < 0 )
    return rc;


  // Check if the number of data bytes to be written is correct.
  if ( tmp_byte != 3 )
    return PCPS_ERR_NBYTES;


  // Write the 3 bytes which are expected:
  rc = _pcps_write_byte( pddev, type );

  if ( rc != PCPS_SUCCESS )
    goto done;


  rc = _pcps_write_byte( pddev, in_cnt );

  if ( rc != PCPS_SUCCESS )
    goto done;


  if ( in_cnt == 0 )
    tmp_byte = out_cnt;
  else
  {
    rc = _pcps_write_byte( pddev, out_cnt );

    if ( rc != PCPS_SUCCESS )
      goto done;


    // Write the input parameters
    p = (uint8_t FAR *) in_buff;
    tmp_byte = in_cnt - 1;

    for ( i = 0; i < tmp_byte; i++ )
    {
      rc = _pcps_write_byte( pddev, *p++ );

      if ( rc < 0 )
        goto done;
    }

    tmp_byte = *p;
  }


  bytes_to_read = 2 + out_cnt;

  if ( bytes_to_read > sizeof( data_read ) )
    bytes_to_read = sizeof( data_read );


  // Write the last byte and read the completion code.
  rc = _pcps_read( pddev, tmp_byte, data_read, (uint8_t) bytes_to_read );

  if ( out_cnt )   //##++ should do some more plausibility checks
    if ( rc == PCPS_SUCCESS )
    {
      _fmemcpy( out_buff, &data_read[2], out_cnt );
    }

done:
  // If an error code has been returned by the I/O function,
  // return that code, otherwise return the completion code
  // read from the board.
  return ( rc < 0 ) ? rc : data_read[0];

}  // pcps_generic_io



/*--------------------------------------------------------------
 * Name:         pcps_read_gps_block()
 *
 * Purpose:      Get a block of data from GPS clock device.
 *               This is a local function which is called
 *               by pcps_read_gps().
 *
 * Input:        pddev         pointer to the device information
 *               data_type     the code assigned to the data type
 *               buffer_size   the size of the buffer
 *               block_num     the number of the block to read
 *               block_size    the size of the block to read
 *
 * Output:       buffer      filled with data
 *
 * Ret value:    PCPS_SUCCESS
 *               PCPS_ERR_TIMEOUT
 *               PCPS_ERR_NBYTES
 *-------------------------------------------------------------*/

static /*HDR*/
short pcps_read_gps_block( PCPS_DDEV *pddev,
                           uint8_t data_type,
                           void FAR *buffer,
                           uint16_t buffer_size,
                           uint8_t block_num,
                           uint8_t block_size )
{
  short rc;
  uint16_t n_bytes;
  uint8_t size_n_bytes;
  uint8_t uc;


  /* Determine which interface buffer size is supported
     and use the appropriate size specification */
  if ( _pcps_ddev_has_gps_data_16( pddev ) )
    size_n_bytes = 2;
  else
  {
    if ( buffer_size > 255 )
      return PCPS_ERR_NBYTES;   // Error ...

    size_n_bytes = 1;
  }

  // Write the command, expect to read one byte.
  rc = _pcps_read_var( pddev, PCPS_READ_GPS_DATA, uc );

  if ( rc != PCPS_SUCCESS )   // Error ...
    return rc;

  if ( uc != 1 ) // The board doesn't expect exactly one more byte
  {
    if ( uc == 0 )
    {
      // The board can't respond now. This may occur if a
      // GPS receiver is still initializing after power-up.
      return PCPS_ERR_NOT_READY;
    }
    else
      return PCPS_ERR_NBYTES;   // Error ...
  }

  // Write the code corresponding to the type of data we
  // want to read, expect to read the expected size.
  n_bytes = 0;
  rc = _pcps_read( pddev, data_type, &n_bytes, size_n_bytes );

  if ( rc != PCPS_SUCCESS )   // Error ...
    return rc;

  if ( n_bytes == 0 )
    return PCPS_ERR_INV_TYPE;

  if ( n_bytes != buffer_size )  // Size of data structure does not match.
    return PCPS_ERR_NBYTES;


  // Write the block number and read n bytes of data.
  rc = _pcps_read( pddev, block_num, buffer, block_size );

  return rc;

}  // pcps_read_gps_block



/*--------------------------------------------------------------
 * Name:         pcps_read_gps()
 *
 * Purpose:      Get a data structure from a GPS clock.
 *
 * Input:        pddev         pointer to the device information
 *               data_type     the code assigned to the data type
 *               buffer_size   the size of the buffer
 *
 * Output:       buffer        filled with data
 *
 * Ret value:    PCPS_SUCCESS
 *               PCPS_ERR_TIMEOUT
 *               PCPS_ERR_NBYTES
 *-------------------------------------------------------------*/

/*HDR*/
short pcps_read_gps( PCPS_DDEV *pddev,
                     uint8_t data_type,
                     void FAR *buffer,
                     uint16_t buffer_size )
{
  uint8_t FAR *p = (uint8_t FAR *) buffer;
  short rc = 0;
  int dt_quot;
  int dt_rem;
  int block_num;


  // Split buffer size to a number of blocks of PCPS_FIFO_SIZE
  // and a number of remaining bytes (less than PCPS_FIFO_SIZE).
  dt_quot = buffer_size / PCPS_FIFO_SIZE;
  dt_rem = buffer_size % PCPS_FIFO_SIZE;

  // Read dt_quot full blocks of data.
  for ( block_num = 0; block_num < dt_quot; block_num++ )
  {
    rc = pcps_read_gps_block( pddev, data_type, p, buffer_size,
                              (uint8_t) block_num, PCPS_FIFO_SIZE );

    if ( rc != PCPS_SUCCESS )   // Error ...
      return rc;

    // Move the destination pointer to the next free byte.
    p += PCPS_FIFO_SIZE;
  }


  // Read dt_rem additional bytes of data.
  if ( dt_rem )
    rc = pcps_read_gps_block( pddev, data_type, p, buffer_size,
                              (uint8_t) block_num, (uint8_t) dt_rem );

  return rc;

}  // pcps_read_gps



/*--------------------------------------------------------------
 * Name:         pcps_write_gps()
 *
 * Purpose:      Write a data structure to a GPS clock.
 *
 * Input:        pddev         pointer to the device information
 *               data_type     the code assigned to the data type
 *               buffer        the data to write
 *               buffer_size   the size of the buffer
 *               read_fnc      function to access the board
 *
 * Output:       --
 *
 * Ret value:    PCPS_SUCCESS
 *               PCPS_ERR_TIMEOUT
 *               PCPS_ERR_NBYTES
 *-------------------------------------------------------------*/

/*HDR*/
short pcps_write_gps( PCPS_DDEV *pddev,
                      uint8_t data_type,
                      const void FAR *buffer,
                      uint16_t buffer_size )
{
  const uint8_t FAR *p = (const uint8_t FAR *) buffer;
  short rc;
  short i;
  uint16_t n_bytes;
  uint8_t size_n_bytes;
  uint8_t uc;


  /* Determine which interface buffer size is supported
     and use the appropriate size specification */
  if ( _pcps_ddev_has_gps_data_16( pddev ) )
    size_n_bytes = 2;
  else
  {
    if ( buffer_size > 255 )
      return PCPS_ERR_NBYTES;   // Error ...

    size_n_bytes = 1;
  }

  // Write the command, expect to read one byte.
  rc = _pcps_read_var( pddev, PCPS_WRITE_GPS_DATA, uc );

  if ( rc != PCPS_SUCCESS )   // Error ...
    return rc;

  if ( uc != 1 )  // The board doesn't expect exactly one more byte.
    return PCPS_ERR_NBYTES;


  // Write the code corresponding to the type of data we
  // want to write, expect to read the expected size.
  n_bytes = 0;
  rc = _pcps_read( pddev, data_type, &n_bytes, size_n_bytes );

  if ( rc != PCPS_SUCCESS )   // Error ...
    return rc;

  if ( n_bytes != buffer_size )  // The board doesn't expect the number
    return PCPS_ERR_NBYTES;      // of bytes we were going to write.


  // Write all bytes but the last one without reading.
  buffer_size--;

  for ( i = 0; i < buffer_size; i++ )
  {
    rc = _pcps_write_byte( pddev, *p++ );

    if ( rc != PCPS_SUCCESS )   // Error ...
      return rc;
  }


  // Write the last byte and read the completion code.
  rc = _pcps_read_var( pddev, *p, n_bytes );

  // If an error code has been returned by read_fnc, return that
  // code, otherwise return the completion code read from the board.
  return rc ? rc : n_bytes;

}  // pcps_write_gps



/*--------------------------------------------------------------
 * Name:         pcps_get_fw_id()
 *
 * Purpose:      This function tries to read the firmware ID
 *               from the board. It should be used to check
 *               if the board is properly installed and can
 *               be accessed without problems.
 *
 * Input:        pddev       pointer to the device information
 *
 * Output:       fw_id       buffer filled with ASCIIZ string
 *
 * Ret value:    PCPS_SUCCESS     no error
 *               PCPS_ERR_TIMEOUT the board is busy for too long
 *-------------------------------------------------------------*/

/*HDR*/
short pcps_get_fw_id( PCPS_DDEV *pddev, PCPS_ID_STR FAR fw_id )
{
  short rc;


  // read first part of the firmware ID
  rc = _pcps_read( pddev, PCPS_GIVE_FW_ID_1, &fw_id[0], PCPS_FIFO_SIZE );

  if ( rc != PCPS_SUCCESS )
    return rc;   // may be timeout


  // read second part of the firmware ID
  rc = _pcps_read( pddev, PCPS_GIVE_FW_ID_2, &fw_id[PCPS_FIFO_SIZE], PCPS_FIFO_SIZE );

  if ( rc != PCPS_SUCCESS )
    return rc;   // may be timeout


  // terminate the string with 0

  fw_id[PCPS_ID_SIZE - 1] = 0;


  return PCPS_SUCCESS;

}  // pcps_get_fw_id



/*--------------------------------------------------------------
 * Name:         pcps_check_id()
 *
 * Purpose:      Check an ASCIIZ string for a valid signature.
 *
 * Input:        pddev       pointer to the device information
 *               ref         the reference signature
 *
 * Output:       --
 *
 * Ret value:    PCPS_SUCCESS      no error
 *               PCPS_ERR_FW_ID    the firmware ID is not valid
 *-------------------------------------------------------------*/

/*HDR*/
short pcps_check_id( PCPS_DDEV *pddev, const char FAR *ref )
{
  // check if the first characters of the string match the reference

  if ( ref )
    if ( _fstrncmp( _pcps_ddev_fw_id( pddev ), ref, _fstrlen( ref ) ) )
      return PCPS_ERR_FW_ID;

  return PCPS_SUCCESS;

}  // pcps_check_id



/*--------------------------------------------------------------
 * Name:         pcps_get_rev_num()
 *
 * Purpose:      Get a version number from an ID string.
 *
 * Input:        idstr     the ID string
 *
 * Output:       --
 *
 * Ret value:    on success: the version number in hex
 *                           (e.g. 0x0270 for version 2.7)
 *               on error: 0
 *-------------------------------------------------------------*/

/*HDR*/
short pcps_get_rev_num( char FAR *idstr )
{
  int i;
  int len = _fstrlen( idstr ) - 2;
  char c;

  uchar rev_num_hi;
  uchar rev_num_lo;

  for ( i = 0; i < len; i++ )
  {
    if ( idstr[i + 1] == '.' )
    {
      rev_num_hi = idstr[i] & 0x0F;
      rev_num_lo = ( idstr[i + 2] & 0x0F ) << 4;

      c = idstr[i + 3];

      if ( c >= '0' && c <= '9' )
        rev_num_lo |= c & 0x0F;

      return ( rev_num_hi << 8 ) | rev_num_lo;
    }
  }

  return 0;

}  // pcps_get_rev_num



/*--------------------------------------------------------------
 * Name:         pcps_read_sernum()
 *
 * Purpose:      This function tries to read the clock's S/N
 *               from the board, if supported by the clock.
 *
 * Input:        pddev       pointer to the device information
 *
 * Output:       pddev       sernum field filled with ASCIIZ
 *
 * Ret value:    PCPS_SUCCESS     no error
 *               other            error
 *-------------------------------------------------------------*/

/*HDR*/
int pcps_read_sernum( PCPS_DDEV *pddev )
{
  int rc = PCPS_SUCCESS;


  memset( pddev->dev.cfg.sernum, 0, sizeof( pddev->dev.cfg.sernum ) );

  // There are different ways to read the clock's S/N.
  // Check which way is supported by the clock, and
  // read the S/N,

  // Read directly. This is supported by newer DCF77 clocks.
  if ( _pcps_ddev_has_sernum( pddev ) )
  {
    rc = _pcps_read( pddev, PCPS_GIVE_SERNUM, pddev->dev.cfg.sernum, PCPS_FIFO_SIZE );

    if ( rc != PCPS_SUCCESS )
    {
      _mbgddmsg_2( MBG_DBG_INIT_DEV, "PCPS read SERNUM %X: rc = %i",
                   _pcps_ddev_dev_id( pddev ), rc );
      goto fail;
    }

    goto done;
  }


  // The S/N is part of the RECEIVER_INFO structure,
  // so use that one, if supported.
  if ( _pcps_ddev_has_receiver_info( pddev ) )
  {
    static_wc RECEIVER_INFO ri;

    rc = _pcps_read_gps_var( pddev, PC_GPS_RECEIVER_INFO, ri );

    if ( rc != PCPS_SUCCESS )
    {
      _mbgddmsg_2( MBG_DBG_INIT_DEV, "PCPS read GPS reciver info %X: rc = %i",
                   _pcps_ddev_dev_id( pddev ), rc );
      goto fail;
    }

    strncpy( pddev->dev.cfg.sernum, ri.sernum,
             sizeof( pddev->dev.cfg.sernum ) );
    goto done;
  }


  // Older GPS clocks store the S/N in an IDENT structure
  // which must be decoded to get the S/N.
  if ( _pcps_ddev_has_ident( pddev ) )
  {
    static_wc IDENT ident;

    rc = _pcps_read_gps_var( pddev, PC_GPS_IDENT, ident );

    #if !defined( MBG_TGT_LINUX )
      assert( sizeof( ident ) < sizeof( pddev->dev.cfg.sernum ) );
    #endif

    if ( rc != PCPS_SUCCESS )
    {
      _mbgddmsg_2( MBG_DBG_INIT_DEV, "PCPS read GPS ident %X: rc = %i",
                   _pcps_ddev_dev_id( pddev ), rc );
      goto fail;
    }

    mbg_gps_ident_decode( _pcps_ddev_sernum( pddev ), &ident );
    goto done;
  }

  // The clock doesn't support a S/N.
  // Assume the rc is still set to PCPS_SUCCESS.
  strcpy( _pcps_ddev_sernum( pddev ), "N/A" );
  goto done;


fail:
  memset( pddev->dev.cfg.sernum, '?', 8 );

done:
  return rc;

}  // pcps_read_sernum



#if _PCPS_USE_RSRCMGR

static /*HDR*/
void pcps_rsrc_alloc_io( PCPS_DDEV *pddev )
{
  ushort decode_width;
  int i;

  if ( _pcps_ddev_is_pci( pddev ) )
    decode_width = PCPS_DECODE_WIDTH_PCI;
  else
    if ( _pcps_ddev_is_mca( pddev ) )
      decode_width = PCPS_DECODE_WIDTH_MCA;
    else
      decode_width = PCPS_DECODE_WIDTH_ISA;


  for ( i = 0; i < N_PCPS_PORT_RSRC; i++ )
  {
    PCPS_PORT_RSRC *p = &_pcps_ddev_port_rsrc( pddev, i );
    ushort rc;

    if ( _pcps_port_rsrc_unused( p ) )
      continue;


    rc = _rsrc_alloc_ports( &pddev->rsrc, p->base, p->num, decode_width );

    // If the resource manager was unable to alloc the resources
    // then the selected range of ports is probably in use
    // by another hardware device and/or driver
    if ( rc )
      _pcps_ddev_set_err_flags( pddev, PCPS_EF_IO_RSRC );
  }

}  // pcps_rsrc_alloc_io



static /*HDR*/
void pcps_rsrc_dealloc_io( PCPS_DDEV *pddev )
{
  int i;

  for ( i = 0; i < N_PCPS_PORT_RSRC; i++ )
  {
    PCPS_PORT_RSRC *p = &_pcps_ddev_port_rsrc( pddev, i );

    if ( _pcps_port_rsrc_unused( p ) )
      continue;

    // clean up if clock not found
    _rsrc_dealloc_ports( &pddev->rsrc.hResource[i], p->base, p->num );
  }

}  // pcps_rsrc_dealloc_io



#ifdef _RSRC_DEV

static /*HDR*/
void pcps_rsrc_register_device( PCPS_DDEV *pddev )
{
  uchar bus_type;
  ushort rc;

  if ( _pcps_ddev_is_pci( pddev ) )
    bus_type = RSRC_BUS_PCI;
  else
    if ( _pcps_ddev_is_mca( pddev ) )
      bus_type = RSRC_BUS_MCA;
    else
      bus_type = RSRC_BUS_ISA;

  rc = rsrc_register_device( &pddev->hDev,
                             &pddev->rsrc,
                             n_ddevs - 1,
                             _pcps_ddev_is_gps( pddev ) ?
                             rsrc_clock_gps : rsrc_clock_dcf77,
                             bus_type );

}  // pcps_rsrc_register_device

#endif  /* def _RSRC_DEV */



/*HDR*/
void pcps_rsrc_release( PCPS_DDEV *pddev )
{
  pddev->read = pcps_read_null;

  pcps_rsrc_dealloc_io( pddev );

}  // pcps_rsrc_release



/*HDR*/
void pcps_rsrc_release_all( void )
{
  int i;

  for ( i = 0; i < n_ddevs; i++ )
    pcps_rsrc_release( &pcps_ddev[i] );

}  // pcps_rsrc_release_all

#endif  /* _PCPS_USE_RSRCMGR */



#if _PCPS_USE_MCA

/*--------------------------------------------------------------
 *  PS31 only:
 *
 *  The scheme below shows the way the bits of the POS 103
 *  configuration byte are mapped to the PS31's programmable
 *  address decoder:
 *
 *   MSB            LSB
 *    |      ||      |
 *    0bbbbbb1000bxxxx  <-- 16 bit port base address (binary)
 *     ||||||    |
 *      \\\\\\   |               b:  configurable bit
 *        \\\\\\ |               x:  don't care bit
 *         |||||||
 *        1bbbbbbb      <--  8 bit configuration byte (POS 103)
 *        |
 *        |
 *    decoder enable bit, always 1 if adapter enabled
 *
 *-------------------------------------------------------------*/

/*--------------------------------------------------------------
 * Convert the code read from PS/2 POS to the port base address.
 *-------------------------------------------------------------*/

/*HDR*/
ushort pcps_port_from_pos( ushort pos )
{
  ushort us = ( ( pos & 0x007E ) << 8 ) | 0x0100;

  if ( pos & 0x0001 )
    us |= 0x0010;

  return us;

}  // pcps_port_from_pos



/*--------------------------------------------------------------
 * Convert the port base address to a PS/2 POS code.
 *-------------------------------------------------------------*/

/*HDR*/
uchar pcps_pos_from_port( ushort port )
{
  uchar uc;


  uc = *( ( (uchar *) (&port) ) + 1 ) & 0x7E;

  if ( port & 0x0010 )
    uc |= 1;

  return uc;

}  // pcps_pos_from_port

#endif  /* _PCPS_USE_MCA */



static /*HDR*/
void check_feature( PCPS_DDEV *pddev, ushort req_rev_num,
                    PCPS_FEATURES flag )
{
  if ( pddev->dev.cfg.fw_rev_num >= req_rev_num )
    pddev->dev.cfg.features |= flag;

}  // check_feature



// The function below takes a bus flag and device ID to search
// the table of known devices for a device which matches the
// given criteria.

/*HDR*/
PCPS_DEV_TYPE *pcps_get_dev_type( int bus_mask, ushort dev_id )
{
  int i;

  for ( i = 0; i < N_PCPS_DEV_TYPE; i++ )
  {
    PCPS_DEV_TYPE *p = &pcps_dev_type[i];

    if ( !( p->bus_flags & bus_mask ) )
      continue;

    if ( p->dev_id == dev_id )
      return p;
  }

  return NULL;

}  // pcps_get_dev_type



/*HDR*/
PCPS_DDEV *pcps_add_new_ddev( void )
{
  PCPS_DDEV *pddev;

  if ( n_ddevs >= PCPS_MAX_DDEVS )
  {
    _mbgddmsg_0( MBG_DBG_INIT_DEV,
                 "Unable to add new device: max count reached" );

    return NULL;
  }


  pddev = &pcps_ddev[n_ddevs];
  n_ddevs++;

  memset( pddev, 0, sizeof( *pddev ) );

  return pddev;

}  // pcps_add_new_ddev



/*HDR*/
void pcps_remove_last_ddev( void )
{
  if ( n_ddevs )
    n_ddevs--;

}  // pcps_remove_last_ddev



static /*HDR*/
void init_ddev_cfg( PCPS_DDEV *pddev )
{
  ushort port_offs;


  #if _PCPS_USE_RSRCMGR
    pcps_rsrc_alloc_io( pddev );
  #endif

  // Determine the I/O port address which must be read to determine
  // the clock's interface status. The address depends on the
  // clock's bus type and interface chip.

  port_offs = _pcps_ddev_is_pci_asic( pddev )
              ? offsetof( PCI_ASIC, status_port )
              : ( _pcps_ddev_is_pci( pddev )
                ? ( AMCC_OP_REG_IMB4 + 3 )
                : 1 );

  pddev->dev.cfg.status_port = _pcps_ddev_port_base( pddev, 0 ) + port_offs;
  pddev->dev.cfg.timeout_clk = PCPS_TIMEOUT_CNT;

}  // init_ddev_cfg



static /*HDR*/
void probe_ddev_cfg( PCPS_DDEV *pddev )
{
  // try to read EPROM ID
  short rc = pcps_get_fw_id( pddev, pddev->dev.cfg.fw_id );

  if ( rc < 0 )
  {
    _mbgddmsg_2( MBG_DBG_INIT_DEV, "PCPS probe device %X: rc = %i",
                 _pcps_ddev_dev_id( pddev ), rc );

    _pcps_ddev_set_err_flags( pddev, PCPS_EF_TIMEOUT );
  }

}  // probe_ddev_cfg



static /*HDR*/
void check_opt_features( PCPS_DDEV *pddev )
{
  if ( _pcps_ddev_has_receiver_info( pddev ) )
  {
    // detect the presence of some optional features at run time
    RECEIVER_INFO rcvr_info;
    int rc;

    rc = _pcps_read_gps_var( pddev, PC_GPS_RECEIVER_INFO, rcvr_info );

    if ( rc == PCPS_SUCCESS )
    {
      if ( rcvr_info.features & GPS_HAS_IRIG_TX )
        pddev->dev.cfg.features |= PCPS_HAS_IRIG_TX;

      if ( rcvr_info.features & GPS_HAS_SYNTH )
        pddev->dev.cfg.features |= PCPS_HAS_SYNTH;
    }
  }

}  // check_opt_features



static /*HDR*/
void finish_ddev_cfg( PCPS_DDEV *pddev )
{
  #ifdef _RSRC_DEV
    pcps_rsrc_register_device( pddev );
  #endif

  // Extract the firmware revision number from the ID string.
  pddev->dev.cfg.fw_rev_num = pcps_get_rev_num( _pcps_ddev_fw_id( pddev ) );

  switch( _pcps_ddev_type_num( pddev ) )
  {
    case PCPS_TYPE_PC31:
    case PCPS_TYPE_PS31_OLD:
    case PCPS_TYPE_PS31:
      pddev->dev.cfg.features = PCPS_FEAT_PC31PS31;
      check_feature( pddev, REV_CAN_SET_TIME_PC31PS31, PCPS_CAN_SET_TIME );
      check_feature( pddev, REV_HAS_SERIAL_PC31PS31, PCPS_HAS_SERIAL );
      check_feature( pddev, REV_HAS_SYNC_TIME_PC31PS31, PCPS_HAS_SYNC_TIME );
      check_feature( pddev, REV_HAS_UTC_OFFS_PC31PS31, PCPS_HAS_UTC_OFFS );
      break;

    case PCPS_TYPE_PC32:
      pddev->dev.cfg.features = PCPS_FEAT_PC32;
      break;

    case PCPS_TYPE_PCI32:
      pddev->dev.cfg.features = PCPS_FEAT_PCI32;
      break;

    case PCPS_TYPE_GPS167PC:
      pddev->dev.cfg.features = PCPS_FEAT_GPS167PC;
      check_feature( pddev, REV_HAS_HR_TIME_GPS167PC, PCPS_HAS_HR_TIME );
      check_feature( pddev, REV_HAS_CABLE_LEN_GPS167PC, PCPS_HAS_CABLE_LEN );
      break;

    case PCPS_TYPE_GPS167PCI:
      pddev->dev.cfg.features = PCPS_FEAT_GPS167PCI;
      check_feature( pddev, REV_HAS_CABLE_LEN_GPS167PCI, PCPS_HAS_CABLE_LEN );
      check_feature( pddev, REV_CAN_CLR_UCAP_BUFF_GPS167PCI, PCPS_CAN_CLR_UCAP_BUFF );
      check_feature( pddev, REV_HAS_UCAP_GPS167PCI, PCPS_HAS_UCAP );
      break;

    case PCPS_TYPE_PCI509:
      pddev->dev.cfg.features = PCPS_FEAT_PCI509;
      break;

    case PCPS_TYPE_GPS168PCI:
      pddev->dev.cfg.features = PCPS_FEAT_GPS168PCI;
      check_feature( pddev, REV_CAN_CLR_UCAP_BUFF_GPS168PCI, PCPS_CAN_CLR_UCAP_BUFF );
      check_feature( pddev, REV_HAS_UCAP_GPS168PCI, PCPS_HAS_UCAP );
      break;

    case PCPS_TYPE_PCI510:
      pddev->dev.cfg.features = PCPS_FEAT_PCI510;
      break;

    case PCPS_TYPE_GPS169PCI:
      pddev->dev.cfg.features = PCPS_FEAT_GPS169PCI;
      check_opt_features( pddev );
      check_feature( pddev, REV_HAS_GPS_DATA_16_GPS169PCI, PCPS_HAS_GPS_DATA_16 );
      break;

    case PCPS_TYPE_TCR510PCI:
      pddev->dev.cfg.features = PCPS_FEAT_TCR510PCI;
      check_feature( pddev, REV_HAS_HR_TIME_TCR510PCI, PCPS_HAS_HR_TIME );
      break;

    case PCPS_TYPE_TCR167PCI:
      pddev->dev.cfg.features = PCPS_FEAT_TCR167PCI;
      check_opt_features( pddev );
      break;

    case PCPS_TYPE_GPS170PCI:
      pddev->dev.cfg.features = PCPS_FEAT_GPS170PCI;
      check_opt_features( pddev );
      break;

    case PCPS_TYPE_PCI511:
      pddev->dev.cfg.features = PCPS_FEAT_PCI511;
      break;

    case PCPS_TYPE_TCR511PCI:
      pddev->dev.cfg.features = PCPS_FEAT_TCR511PCI;
      break;

  }  // switch

  #ifndef MBG_TGT_OS2
    // Function strstr not supported at kernel level,
    // but this is not required, either.
    if ( strstr( _pcps_ddev_fw_id( pddev ), "CERN" ) != NULL )
      pddev->dev.cfg.features |= PCPS_HAS_EVENT_TIME;
  #endif

  pcps_read_sernum( pddev );

}  // finish_ddev_cfg



/*--------------------------------------------------------------
 * PCI functions
 *-------------------------------------------------------------*/

#if _PCPS_USE_PCIMGR

/*HDR*/
int pcps_pnp_add_device( PCPS_DDEV *pddev, int bus_flags, ushort dev_id )
{
  // First check if we really support the device to be added.
  PCPS_DEV_TYPE *pdt = pcps_get_dev_type( bus_flags, dev_id );

  if ( pdt == 0 )
  {
    _mbgddmsg_1( MBG_DBG_INIT_DEV, "PCPS add PNP device %X: unknown type",
                 dev_id );

    return -1;
  }


  pddev->dev.type = *pdt;

  _mbgddmsg_2( MBG_DBG_INIT_DEV, "PCPS add PNP device %X: found %s",
               dev_id, pdt->name );

  return PCPS_SUCCESS;

}  // pcps_pnp_add_device

#endif /* _PCPS_USE_PCIMGR */



#if _PCPS_USE_PCI

/*HDR*/
int pcps_pnp_start_device( PCPS_DDEV *pddev,
                           PCPS_BUS_NUM bus_num,
                           PCPS_SLOT_NUM dev_fnc_num,
                           PCPS_RSRC *prsrc )
{
  int i;

  _mbgddmsg_1( MBG_DBG_INIT_DEV, "PCPS start PNP device %X",
               _pcps_ddev_dev_id( pddev ) );

  pddev->dev.cfg.bus_num = bus_num;
  pddev->dev.cfg.slot_num = dev_fnc_num;

  for ( i = 0; i < N_PCPS_PORT_RSRC; i++ )
    pddev->dev.cfg.port[i] = prsrc->port[i];

  pddev->dev.cfg.irq_num = prsrc->irq_num;


  if ( _pcps_ddev_chk_err_flags( pddev, PCPS_EF_IO_INIT | PCPS_EF_IO_ENB ) )
  {
    pddev->read = pcps_read_null;
    return -1;
  }


  if ( _pcps_ddev_is_pci_asic( pddev ) )
    pddev->read = pcps_read_asic;
  else
    if ( _pcps_ddev_is_pci_s5920( pddev ) )
      pddev->read = pcps_read_amcc_s5920;
    else
      pddev->read = pcps_read_amcc_s5933;

  init_ddev_cfg( pddev );
  probe_ddev_cfg( pddev );
  finish_ddev_cfg( pddev );

  return 0;

}  // pcps_pnp_start_device

#endif /* _PCPS_USE_PCI */



#if _PCPS_USE_PCIBIOS

static /*HDR*/
PCPS_ERR_FLAGS pcps_read_pci_rsrc( PCPS_BUS_NUM bus_num,
                                   PCPS_SLOT_NUM dev_fnc_num,
                                   PCPS_RSRC *prsrc,
                                   PCPS_BUS_FLAGS bus_flags )
{
  PCPS_ERR_FLAGS err_flags = 0;
  PCPS_PORT_RSRC *pprsrc;
  uchar irq;
  short rc;
  PCI_DWORD dw;

  // Clear resources
  memset( prsrc, 0, sizeof( *prsrc ) );
  prsrc->irq_num = -1;

  // Read base address 0
  rc = _mbg_pci_read_cfg_dword( bus_num, dev_fnc_num,
                                PCI_CS_BASE_ADDRESS_0, &dw );

  // LSB is set to '1' if configured as I/O address space.
  // Base address 0 _must_ always be configured as I/O.
  if ( !( rc == PCI_SUCCESS && ( dw & 0x0001 ) ) )
  {
    err_flags |= PCPS_EF_IO_CFG;
    goto done;
  }


  if ( dw & 0xFFFF0000UL || ( ( dw & ~0x0001 ) == 0 ) )
  {
    // The PCI interface chip is not initialized. This
    // should occur ONLY at the first-time installation
    // at the factory.
    err_flags |= PCPS_EF_IO_INIT;
    goto done;
  }

  // Save base address range 0.
  pprsrc = &prsrc->port[0];
  pprsrc->base = (ushort) dw & ~0x0001;

  pprsrc->num = ( bus_flags & PCPS_BUS_PCI_ASIC ) ?
                  sizeof( PCI_ASIC ) :
                  ( ( bus_flags & PCPS_BUS_PCI_S5920 ) ?
                    AMCC_OP_REG_RANGE_S5920 :
                    AMCC_OP_REG_RANGE_S5933
                  );

  // Base address 1 must be configured
  // if S5920 chip is used.
  pprsrc = &prsrc->port[1];

  if ( bus_flags & PCPS_BUS_PCI_S5920 )
  {
    // Read base address 1.
    rc = _mbg_pci_read_cfg_dword( bus_num, dev_fnc_num,
                                  PCI_CS_BASE_ADDRESS_1, &dw );

    // LSB is set to '1' if configured as I/O address space.
    if ( rc == PCI_SUCCESS && ( dw & 0x0001 ) != 0 )
    {
      // Save base address range 1.
      pprsrc->base = (ushort) dw & ~0x0001;
      pprsrc->num = 16;
    }
    else
    {
      err_flags |= PCPS_EF_IO_CFG;
      goto done;
    }
  }

  // Read the interrupt line assigned to the clock.
  // The standard functions, however, don't use any
  // interrupt.
  rc = _mbg_pci_read_cfg_byte( bus_num, dev_fnc_num,
                               PCI_CS_INTERRUPT_LINE, &irq );

  if ( rc == PCI_SUCCESS )
    prsrc->irq_num = irq;

done:
  return err_flags;

}  // pcps_read_pci_rsrc



static /*HDR*/
PCPS_ERR_FLAGS pcps_enable_pci_dev( PCPS_BUS_NUM bus_num,
                                    PCPS_SLOT_NUM dev_fnc_num )
{
  PCPS_ERR_FLAGS err_flags = 0;
  ushort pci_command;
  int rc;


  // If the option "PNP OS installed" is set to "YES" in the
  // PC's BIOS setup, then I/O access to the board may still
  // be disabled, so check if the clock is enabled and enable
  // access to the board, if nessessary.
  rc = _mbg_pci_read_cfg_word( bus_num, dev_fnc_num,
                               PCI_CS_COMMAND, &pci_command );

  if ( !( pci_command & PCI_CMD_ENB_IO_ACC ) )
  {
    pci_command = PCI_CMD_ENB_IO_ACC;
    rc = _mbg_pci_write_cfg_word( bus_num, dev_fnc_num,
                                  PCI_CS_COMMAND, pci_command );
    if ( rc != PCI_SUCCESS )
    {
      err_flags |= PCPS_EF_IO_ENB;

      _mbgddmsg_1( MBG_DBG_INIT_DEV,
                   "PCI enable device returned %d", rc );
    }
  }

  return err_flags;

}  // pcps_enable_pci_dev



/*HDR*/
void pcps_setup_pci_dev( PCPS_DDEV *pddev,
                         PCPS_BUS_NUM bus_num,
                         PCPS_SLOT_NUM dev_fnc_num )
{
  static_wc PCPS_RSRC pci_rsrc;
  PCPS_ERR_FLAGS err_flags;

  err_flags = pcps_read_pci_rsrc( bus_num, dev_fnc_num, &pci_rsrc,
                                  _pcps_ddev_bus_flags( pddev ) );

  if ( !( err_flags & PCPS_EF_IO_INIT ) )
    err_flags |= pcps_enable_pci_dev( bus_num, dev_fnc_num );

  if ( err_flags )
    _pcps_ddev_set_err_flags( pddev, err_flags );

  pcps_pnp_start_device( pddev, bus_num, dev_fnc_num, &pci_rsrc );

}  // pcps_setup_pci_dev



/*HDR*/
void pcps_detect_pci_clocks( ushort vendor_id,
                        PCPS_DEV_TYPE dev_type[],
                        int n_dev_types )
{
  #if defined( MBG_TGT_QNX )
    #if defined( MBG_TGT_QNX_NTO )
      unsigned int pci_handle;  // specific to QNX Neutrino
    #endif
    unsigned int pci_hardware_mechanism;
    unsigned int pci_last_bus_number;
    unsigned int pci_interface_level_version;
  #elif defined( MBG_TGT_LINUX )
    // not yet supported/used
  #else
    uchar pci_hardware_mechanism;
    uchar pci_last_bus_number;
    ushort pci_interface_level_version;
  #endif
  ushort type_idx;
  int rc;


  #ifdef _mbg_pci_fnc_init
    rc = _mbg_pci_fnc_init();

    if ( rc != PCI_SUCCESS )
      return;
  #endif


  // See if PCI BIOS is installed on the machine.
  rc = _mbg_pci_find_bios( &pci_hardware_mechanism,
                           &pci_interface_level_version,
                           &pci_last_bus_number
                         );

  if ( rc == PCI_SUCCESS )
  {
    // PCI BIOS is installed, now try to find a PCI clock with
    // known ID (the list is terminated with a ID of 0).
    for ( type_idx = 0; type_idx < n_dev_types; type_idx++ )
    {
      ushort dev_idx;
      PCPS_DEV_TYPE *p = &dev_type[type_idx];

      if ( !( p->bus_flags & PCPS_BUS_PCI ) )
        continue;


      for ( dev_idx = 0; ; dev_idx++ )
      {
        PCPS_DDEV *pddev;
        #if defined( MBG_TGT_QNX )
          unsigned bus_num;
          unsigned dev_fnc_num;
        #else
          uchar bus_num;
          uchar dev_fnc_num;
        #endif

        rc = _mbg_pci_find_device( p->dev_id, vendor_id,
                                   dev_idx, &bus_num, &dev_fnc_num );

        if ( rc != PCI_SUCCESS )
          break;  // go to try next device ID


        // New device found, try to add to list.
        pddev = pcps_add_new_ddev();

        if ( pddev )  // Setup only if successful.
        {
          #if _PCPS_USE_PCIMGR  //##++
          {
            // This is used to test the PNP functions in a
            // non-PNP environment.

            PCPS_RSRC pci_rsrc;
            PCPS_ERR_FLAGS err_flags;

            pcps_pnp_add_device( pddev, PCPS_BUS_PCI, p->dev_id );

            err_flags = pcps_read_pci_rsrc( bus_num, dev_fnc_num, &pci_rsrc,
                                            _pcps_ddev_bus_flags( pddev ) );

            pcps_pnp_start_device( pddev, bus_num, dev_fnc_num, &pci_rsrc );
          }
          #else
            pddev->dev.type = *p;
            pcps_setup_pci_dev( pddev, bus_num, dev_fnc_num );
          #endif

          #if !_ACCEPT_UNINITD_CLOCKS
            if ( pddev->dev.cfg.err_flags )
            {
              _mbgddmsg_1( MBG_DBG_INIT_DEV,
                           "Remove PCI device: err_flags " FMT_08X "h",
                           (ulong) pddev->dev.cfg.err_flags );

              pcps_remove_last_ddev();
            }
          #endif
        }
      }
    }
  }

  #ifdef _mbg_pci_fnc_deinit
    _mbg_pci_fnc_deinit();
  #endif

}  // pcps_detect_pci_clocks

#endif  // _PCPS_USE_PCIBIOS



/*--------------------------------------------------------------
 * Try to detect a MCA clock
 *-------------------------------------------------------------*/

#if _PCPS_USE_MCA

static /*HDR*/
void detect_mca_clocks( void )
{
  short rc;
  ushort type_idx;


  rc = mca_fnc_init();

  if ( rc != MCA_SUCCESS )
    return;


  // MCA is installed, now try to find a MCA clock with
  // known ID.
  for ( type_idx = 0; type_idx < N_PCPS_DEV_TYPE; type_idx++ )
  {
    static_wc MCA_POS_DATA pos_data;
    PCPS_DDEV *pddev;
    PCPS_DEV_TYPE *p = &pcps_dev_type[type_idx];

    PCPS_PORT_RSRC *pprsrc;
    static_wc uchar slot_num;   // the slot in which the board is installed


    if ( !( p->bus_flags & PCPS_BUS_MCA ) )
      continue;


    rc = mca_find_adapter( p->dev_id, &slot_num, &pos_data );

    if ( rc != MCA_SUCCESS )
      continue;


    // New device found, try to add to list.
    pddev = pcps_add_new_ddev();

    if ( pddev )  // Setup only if successful.
    {
      pddev->dev.type = *p;  //##++
      pddev->dev.cfg.slot_num = slot_num;

      pprsrc = &_pcps_ddev_port_rsrc( pddev, 0 );
      pprsrc->base = pcps_port_from_pos( pos_data.pos_103 );
      pprsrc->num = PCPS_NUM_PORTS_MCA;

      pddev->read = pcps_read_std;

      #if 0  //##++
        // Try to read the interrupt line assigned to the clock.
        // The standard functions, however, don't use any interrupt.
        rc = _mbg_pci_read_cfg_byte( bus_num, slot_num,
                                     PCI_CS_INTERRUPT_LINE, &irq );
        pddev->info.cfg.irq_num = ( rc == PCI_SUCCESS ) ? irq : -1;
      #endif

      init_ddev_cfg( pddev );
      probe_ddev_cfg( pddev );
      finish_ddev_cfg( pddev );
    }
  }

  mca_fnc_deinit();

}  // detect_mca_clocks

#endif



/*--------------------------------------------------------------
 * Try to detect ISA clocks
 *-------------------------------------------------------------*/

static /*HDR*/
void detect_isa_clocks( int isa_ports[PCPS_MAX_ISA_CARDS],
                        int isa_irqs[PCPS_MAX_ISA_CARDS] )
{
  int *p_port = isa_ports;
  int *p_irq = isa_irqs;
  PCPS_DDEV *pddev;
  int i;

  if ( p_port == NULL )   // No list has been passed
    return;                  // so don't try to detect ISA clocks.


  for( i = 0; i < PCPS_MAX_ISA_CARDS;
       i++, p_port++, p_irq ? ( p_irq++ ) : p_irq )
  {
    PCPS_PORT_RSRC *pprsrc;
    ushort dev_type;
    ushort irq_num;

    if ( *p_port == 0 )
      continue;

    irq_num = p_irq ? *p_irq : 0;

    _mbgddmsg_2( MBG_DBG_INIT_ISA,
                 "Check ISA device at port " FMT_03X "h, irq %d",
                 *p_port, irq_num );

    // Assume ISA device is available,
    // but clock type is unknown, yet.
    pddev = pcps_add_new_ddev();

    if ( pddev )  // Setup only if successfull.
    {
      // Set up basic cfg for ISA devices.
      pprsrc = &_pcps_ddev_port_rsrc( pddev, 0 );
      pprsrc->base = (ushort) *p_port;
      pprsrc->num = PCPS_NUM_PORTS_ISA;

      pddev->read = pcps_read_std;

      // Init the device structure. This includes registration
      // of I/O ports with the OS's resource manager (if supported),
      // and reading the firmware ID.
      init_ddev_cfg( pddev );
      probe_ddev_cfg( pddev );

      // If an error has occurred, then remove the last
      // device from the list and try next.
      if ( pddev->dev.cfg.err_flags )
      {
        // clean up if clock not found
        #if _PCPS_USE_RSRCMGR
          pcps_rsrc_dealloc_io( pddev );
        #endif

        _mbgddmsg_1( MBG_DBG_INIT_DEV,
                     "ISA device not found: err_flags " FMT_08X "h",
                     (ulong) pddev->dev.cfg.err_flags );

        pcps_remove_last_ddev();
        continue;
      }


      // find out which type of ISA clock we have found
      // check EPROM ID to find out which kind of clock is installed
      if ( pcps_check_id( pddev, fw_id_ref_gps ) == PCPS_SUCCESS )
        dev_type = PCPS_TYPE_GPS167PC;
      else
        if ( pcps_check_id( pddev, fw_id_ref_pcps ) == PCPS_SUCCESS )
        {
          // clock is a PC31, or a PC32 if it has signature code
          // if no support for MCA has been compiled in, it may even
          // be a PS31 which is software compatible with a PC31
          dev_type =
            ( inpw( (ushort) ( _pcps_ddev_port_base( pddev, 0 ) + 2 ) )
              == pcps_dev_type[PCPS_TYPE_PC32].dev_id ) ?
                 PCPS_TYPE_PC32 : PCPS_TYPE_PC31;
        }
        else
        {
          _pcps_ddev_set_err_flags( pddev, PCPS_EF_INV_EPROM_ID );
          continue;
        }

      pddev->dev.type = pcps_dev_type[dev_type];

      if ( irq_num )
        pddev->dev.cfg.irq_num = irq_num;


      // finish setup of device configuration
      finish_ddev_cfg( pddev );
    }
  }

}  // detect_isa_clocks



/*--------------------------------------------------------------
 * Try to detect any plug-in radio clock. If a DOS TSR is
 * installed, be sure it is disabled (BUSY flag set) when
 * this function is called.
 *-------------------------------------------------------------*/

/*HDR*/
void pcps_detect_clocks( int isa_ports[PCPS_MAX_ISA_CARDS], int isa_irqs[PCPS_MAX_ISA_CARDS] )
{
  if ( n_ddevs )
    return;    // has probably been called before


  #ifdef _RSRC_DRVR
    // register driver and init resource manager, if required
    rsrc_register_driver();
  #endif

  #if _PCPS_USE_PCIMGR
    #if defined MBG_TGT_LINUX
      register_pci_devices();
    #endif
  #elif _PCPS_USE_PCIBIOS
    pcps_detect_pci_clocks( PCI_VENDOR_MEINBERG, pcps_dev_type,
                            N_PCPS_DEV_TYPE );
  #endif

  #if _PCPS_USE_MCA
    detect_mca_clocks();
  #endif

  detect_isa_clocks( isa_ports, isa_irqs );

}  // pcps_detect_clocks



